"""
Specialized Methods Module

This module provides specialized ecological analysis methods including phylogenetic diversity,
metacommunity analysis, food web analysis, and other advanced ecological techniques.

Copyright (c) 2025 Mohamed Z. Hatim
"""

import numpy as np
import pandas as pd
import warnings
from typing import Dict, List, Optional, Tuple, Union, Any
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.cluster.hierarchy import linkage, dendrogram
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
import seaborn as sns

try:
    from scipy.sparse import csr_matrix
    from scipy.sparse.csgraph import minimum_spanning_tree, dijkstra
    GRAPH_METHODS_AVAILABLE = True
except ImportError:
    GRAPH_METHODS_AVAILABLE = False
    warnings.warn("Scipy graph methods not available, some network analyses will be limited")

try:
    import networkx as nx
    NETWORKX_AVAILABLE = True
except ImportError:
    NETWORKX_AVAILABLE = False
    warnings.warn("NetworkX not available, advanced network analyses will be limited")


class PhylogeneticDiversityAnalyzer:
    """
    Phylogenetic diversity analysis for vegetation communities.
    
    Calculates phylogenetic diversity metrics using phylogenetic trees
    and community composition data.
    """
    
    def __init__(self, random_state: int = 42):
        """Initialize PhylogeneticDiversityAnalyzer."""
        self.random_state = random_state
        self.phylo_tree = None
        self.distance_matrix = None
        
    def load_phylogeny(self, phylo_data: Union[pd.DataFrame, Dict[str, Any]], 
                      format: str = 'distance_matrix') -> None:
        """
        Load phylogenetic data.
        
        Parameters
        ----------
        phylo_data : Union[pd.DataFrame, Dict[str, Any]]
            Phylogenetic data (distance matrix, tree, etc.)
        format : str, optional
            Format of phylogenetic data ('distance_matrix', 'newick'), by default 'distance_matrix'
        """
        if format == 'distance_matrix':
            if isinstance(phylo_data, pd.DataFrame):
                self.distance_matrix = phylo_data
            else:
                raise ValueError("Distance matrix must be a DataFrame")
        elif format == 'newick':
# Copyright (c) 2025 Mohamed Z. Hatim
            warnings.warn("Newick format support is simplified. Use specialized phylogenetic libraries for full support.")
            self.phylo_tree = phylo_data
        else:
            raise ValueError(f"Unknown phylogenetic data format: {format}")
    
    def calculate_phylogenetic_diversity(self, 
                                       community_data: pd.DataFrame,
                                       metrics: List[str] = None) -> Dict[str, Any]:
        """
        Calculate phylogenetic diversity metrics.
        
        Parameters
        ----------
        community_data : pd.DataFrame
            Community composition data (sites x species)
        metrics : List[str], optional
            Phylogenetic diversity metrics to calculate
            
        Returns
        -------
        Dict[str, Any]
            Phylogenetic diversity results
        """
        if metrics is None:
            metrics = ['pd', 'mpd', 'mntd', 'nri', 'nti']
        
        if self.distance_matrix is None:
            raise ValueError("Phylogenetic data not loaded. Use load_phylogeny() first.")
        
# Copyright (c) 2025 Mohamed Z. Hatim
        common_species = list(set(community_data.columns) & set(self.distance_matrix.index))
        if len(common_species) == 0:
            raise ValueError("No common species between community data and phylogeny")
        
        community_subset = community_data[common_species]
        phylo_subset = self.distance_matrix.loc[common_species, common_species]
        
        results = {
            'metrics': {},
            'species_used': common_species,
            'n_species': len(common_species)
        }
        
# Copyright (c) 2025 Mohamed Z. Hatim
        site_results = {}
        for site in community_subset.index:
            site_composition = community_subset.loc[site]
            present_species = site_composition[site_composition > 0].index.tolist()
            
            if len(present_species) < 2:
# Copyright (c) 2025 Mohamed Z. Hatim
                continue
            
            site_phylo = phylo_subset.loc[present_species, present_species]
            abundances = site_composition.loc[present_species]
            
            site_metrics = {}
            
# Copyright (c) 2025 Mohamed Z. Hatim
            if 'pd' in metrics:
                site_metrics['pd'] = self._calculate_faith_pd(site_phylo, abundances)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            if 'mpd' in metrics:
                site_metrics['mpd'] = self._calculate_mpd(site_phylo, abundances)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            if 'mntd' in metrics:
                site_metrics['mntd'] = self._calculate_mntd(site_phylo, abundances)
            
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
            if 'nri' in metrics or 'nti' in metrics:
                null_mpd, null_mntd = self._calculate_null_phylo_metrics(
                    phylo_subset, len(present_species), n_iterations=99
                )
                
                if 'nri' in metrics:
                    site_metrics['nri'] = (null_mpd['mean'] - site_metrics.get('mpd', 0)) / null_mpd['std'] if null_mpd['std'] > 0 else 0
                if 'nti' in metrics:
                    site_metrics['nti'] = (null_mntd['mean'] - site_metrics.get('mntd', 0)) / null_mntd['std'] if null_mntd['std'] > 0 else 0
            
            site_results[site] = site_metrics
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if site_results:
            results['metrics'] = pd.DataFrame(site_results).T
        
        return results
    
    def _calculate_faith_pd(self, phylo_distances: pd.DataFrame, 
                          abundances: pd.Series) -> float:
        """Calculate Faith's Phylogenetic Diversity."""
        if len(phylo_distances) < 2:
            return 0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if GRAPH_METHODS_AVAILABLE:
# Copyright (c) 2025 Mohamed Z. Hatim
            distance_matrix = phylo_distances.values
            mst = minimum_spanning_tree(distance_matrix).toarray()
            pd_value = np.sum(mst)
        else:
# Copyright (c) 2025 Mohamed Z. Hatim
            distances = phylo_distances.values
            unique_distances = distances[np.triu_indices(len(distances), k=1)]
            pd_value = np.sum(unique_distances) / len(unique_distances)
        
        return pd_value
    
    def _calculate_mpd(self, phylo_distances: pd.DataFrame, 
                      abundances: pd.Series) -> float:
        """Calculate Mean Pairwise Distance."""
        distances = phylo_distances.values
        weights = abundances.values
        
# Copyright (c) 2025 Mohamed Z. Hatim
        total_weight = 0
        weighted_distance = 0
        
        for i in range(len(distances)):
            for j in range(i + 1, len(distances)):
                weight = weights[i] * weights[j]
                weighted_distance += distances[i, j] * weight
                total_weight += weight
        
        return weighted_distance / total_weight if total_weight > 0 else 0
    
    def _calculate_mntd(self, phylo_distances: pd.DataFrame, 
                       abundances: pd.Series) -> float:
        """Calculate Mean Nearest Taxon Distance."""
        distances = phylo_distances.values
        np.fill_diagonal(distances, np.inf)  # Exclude self-distances
        
# Copyright (c) 2025 Mohamed Z. Hatim
        nearest_distances = []
        for i in range(len(distances)):
            min_dist = np.min(distances[i, :])
            if np.isfinite(min_dist):
                nearest_distances.append(min_dist)
        
        return np.mean(nearest_distances) if nearest_distances else 0
    
    def _calculate_null_phylo_metrics(self, phylo_matrix: pd.DataFrame, 
                                    n_species: int, 
                                    n_iterations: int = 99) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Calculate null model expectations for phylogenetic metrics."""
        null_mpd_values = []
        null_mntd_values = []
        
        for _ in range(n_iterations):
# Copyright (c) 2025 Mohamed Z. Hatim
            sampled_species = np.random.choice(
                phylo_matrix.index, size=n_species, replace=False
            )
            null_phylo = phylo_matrix.loc[sampled_species, sampled_species]
            null_abundances = pd.Series(1, index=sampled_species)  # Equal abundances
            
            null_mpd = self._calculate_mpd(null_phylo, null_abundances)
            null_mntd = self._calculate_mntd(null_phylo, null_abundances)
            
            null_mpd_values.append(null_mpd)
            null_mntd_values.append(null_mntd)
        
        return (
            {'mean': np.mean(null_mpd_values), 'std': np.std(null_mpd_values)},
            {'mean': np.mean(null_mntd_values), 'std': np.std(null_mntd_values)}
        )


class MetacommunityAnalyzer:
    """
    Metacommunity analysis for studying species distributions across multiple sites.
    """
    
    def __init__(self, random_state: int = 42):
        """Initialize MetacommunityAnalyzer."""
        self.random_state = random_state
        
    def elements_of_metacommunity_structure(self, 
                                          community_data: pd.DataFrame,
                                          method: str = 'reciprocal_averaging') -> Dict[str, Any]:
        """
        Analyze Elements of Metacommunity Structure (EMS).
        
        Parameters
        ----------
        community_data : pd.DataFrame
            Community composition data (sites x species)
        method : str, optional
            Ordination method to use, by default 'reciprocal_averaging'
            
        Returns
        -------
        Dict[str, Any]
            EMS analysis results
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        pa_data = (community_data > 0).astype(int)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        coherence = self._calculate_coherence(pa_data)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        turnover = self._calculate_turnover(pa_data)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        boundary_clumping = self._calculate_boundary_clumping(pa_data)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        structure_type = self._classify_metacommunity_structure(
            coherence, turnover, boundary_clumping
        )
        
        return {
            'coherence': coherence,
            'turnover': turnover, 
            'boundary_clumping': boundary_clumping,
            'structure_type': structure_type,
            'interpretation': self._interpret_structure(structure_type)
        }
    
    def _calculate_coherence(self, pa_data: pd.DataFrame) -> Dict[str, Any]:
        """Calculate species coherence."""
        n_sites, n_species = pa_data.shape
        
# Copyright (c) 2025 Mohamed Z. Hatim
        embedded_absences = 0
        total_possible = 0
        
        for species in pa_data.columns:
            presence_sites = pa_data[pa_data[species] == 1].index
            if len(presence_sites) < 2:
                continue
            
# Copyright (c) 2025 Mohamed Z. Hatim
            site_richness = pa_data.sum(axis=1)
            sorted_sites = site_richness.sort_values(ascending=False).index
            
# Copyright (c) 2025 Mohamed Z. Hatim
            presence_positions = [i for i, site in enumerate(sorted_sites) if site in presence_sites]
            if len(presence_positions) < 2:
                continue
            
            first_pos = min(presence_positions)
            last_pos = max(presence_positions)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            for pos in range(first_pos + 1, last_pos):
                site = sorted_sites[pos]
                if pa_data.loc[site, species] == 0:
                    embedded_absences += 1
            
            total_possible += last_pos - first_pos - 1
        
        coherence_value = 1 - (embedded_absences / total_possible) if total_possible > 0 else 1
        
        return {
            'value': coherence_value,
            'embedded_absences': embedded_absences,
            'total_possible': total_possible
        }
    
    def _calculate_turnover(self, pa_data: pd.DataFrame) -> Dict[str, Any]:
        """Calculate species turnover."""
# Copyright (c) 2025 Mohamed Z. Hatim
        site_richness = pa_data.sum(axis=1)
        total_richness = (pa_data.sum(axis=0) > 0).sum()
        mean_alpha = site_richness.mean()
        
        beta_w = (total_richness / mean_alpha) - 1
        
# Copyright (c) 2025 Mohamed Z. Hatim
        pairwise_turnover = []
        for i, site1 in enumerate(pa_data.index):
            for site2 in pa_data.index[i+1:]:
                shared = (pa_data.loc[site1] & pa_data.loc[site2]).sum()
                unique1 = (pa_data.loc[site1] & ~pa_data.loc[site2]).sum()
                unique2 = (~pa_data.loc[site1] & pa_data.loc[site2]).sum()
                
# Copyright (c) 2025 Mohamed Z. Hatim
                turnover = (unique1 + unique2) / (2 * shared + unique1 + unique2) if (shared + unique1 + unique2) > 0 else 0
                pairwise_turnover.append(turnover)
        
        return {
            'whittaker_beta': beta_w,
            'mean_pairwise_turnover': np.mean(pairwise_turnover),
            'pairwise_turnover': pairwise_turnover
        }
    
    def _calculate_boundary_clumping(self, pa_data: pd.DataFrame) -> Dict[str, Any]:
        """Calculate boundary clumping."""
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
        
        site_richness = pa_data.sum(axis=1).sort_values(ascending=False)
        species_ranges = {}
        
        for species in pa_data.columns:
            presence_sites = pa_data[pa_data[species] == 1].index
            if len(presence_sites) == 0:
                continue
            
# Copyright (c) 2025 Mohamed Z. Hatim
            positions = [list(site_richness.index).index(site) for site in presence_sites]
            range_size = max(positions) - min(positions) + 1
            species_ranges[species] = range_size / len(presence_sites)
        
        mean_clumping = np.mean(list(species_ranges.values())) if species_ranges else 1
        
        return {
            'mean_clumping': mean_clumping,
            'species_clumping': species_ranges
        }
    
    def _classify_metacommunity_structure(self, coherence: Dict, turnover: Dict, 
                                        boundary_clumping: Dict) -> str:
        """Classify metacommunity structure type."""
        coherent = coherence['value'] > 0.5
        high_turnover = turnover['whittaker_beta'] > 1.5
        clumped = boundary_clumping['mean_clumping'] < 0.5
        
        if not coherent:
            return "random"
        elif coherent and not high_turnover:
            return "nested"
        elif coherent and high_turnover and not clumped:
            return "evenly_spaced"
        elif coherent and high_turnover and clumped:
            return "clumped"
        else:
            return "intermediate"
    
    def _interpret_structure(self, structure_type: str) -> str:
        """Provide interpretation of metacommunity structure."""
        interpretations = {
            "random": "Species distributions are random with respect to environmental gradients",
            "nested": "Species-poor sites are subsets of species-rich sites",
            "evenly_spaced": "Species have non-overlapping, evenly distributed ranges",
            "clumped": "Species ranges are clustered, suggesting habitat specialization",
            "intermediate": "Intermediate pattern between major structure types"
        }
        return interpretations.get(structure_type, "Unknown structure type")


class NetworkAnalyzer:
    """
    Network analysis for ecological interactions and co-occurrence patterns.
    """
    
    def __init__(self, random_state: int = 42):
        """Initialize NetworkAnalyzer."""
        self.random_state = random_state
        
    def build_cooccurrence_network(self, 
                                 community_data: pd.DataFrame,
                                 method: str = 'correlation',
                                 threshold: float = 0.5) -> Dict[str, Any]:
        """
        Build species co-occurrence network.
        
        Parameters
        ----------
        community_data : pd.DataFrame
            Community composition data (sites x species)
        method : str, optional
            Method for calculating associations, by default 'correlation'
        threshold : float, optional
            Threshold for network edges, by default 0.5
            
        Returns
        -------
        Dict[str, Any]
            Network analysis results
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        if method == 'correlation':
            associations = community_data.corr()
        elif method == 'jaccard':
            associations = self._calculate_jaccard_matrix(community_data)
        else:
            raise ValueError(f"Unknown method: {method}")
        
# Copyright (c) 2025 Mohamed Z. Hatim
        np.fill_diagonal(associations.values, 0)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        adjacency = (np.abs(associations) >= threshold).astype(int)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        network_props = self._calculate_network_properties(adjacency, associations)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        network_obj = None
        if NETWORKX_AVAILABLE:
            network_obj = self._create_networkx_graph(associations, threshold)
            network_props.update(self._calculate_networkx_properties(network_obj))
        
        return {
            'associations': associations,
            'adjacency_matrix': pd.DataFrame(adjacency, 
                                           index=associations.index, 
                                           columns=associations.columns),
            'network_properties': network_props,
            'network_object': network_obj,
            'method': method,
            'threshold': threshold
        }
    
    def _calculate_jaccard_matrix(self, community_data: pd.DataFrame) -> pd.DataFrame:
        """Calculate Jaccard similarity matrix."""
        pa_data = (community_data > 0).astype(int)
        n_species = pa_data.shape[1]
        jaccard_matrix = np.zeros((n_species, n_species))
        
        for i in range(n_species):
            for j in range(n_species):
                if i == j:
                    jaccard_matrix[i, j] = 1.0
                    continue
                
                species_i = pa_data.iloc[:, i]
                species_j = pa_data.iloc[:, j]
                
                intersection = np.sum(species_i & species_j)
                union = np.sum(species_i | species_j)
                
                jaccard_matrix[i, j] = intersection / union if union > 0 else 0
        
        return pd.DataFrame(jaccard_matrix, 
                          index=community_data.columns, 
                          columns=community_data.columns)
    
    def _calculate_network_properties(self, adjacency: Union[np.ndarray, pd.DataFrame], 
                                    associations: pd.DataFrame) -> Dict[str, Any]:
        """Calculate basic network properties."""
# Copyright (c) 2025 Mohamed Z. Hatim
        if isinstance(adjacency, pd.DataFrame):
            adjacency_values = adjacency.values
        else:
            adjacency_values = adjacency
            
        n_nodes = adjacency_values.shape[0]
        n_edges = np.sum(adjacency_values) // 2  # Undirected network
        
# Copyright (c) 2025 Mohamed Z. Hatim
        degrees = np.sum(adjacency_values, axis=1)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        max_edges = n_nodes * (n_nodes - 1) // 2
        density = n_edges / max_edges if max_edges > 0 else 0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if GRAPH_METHODS_AVAILABLE:
# Copyright (c) 2025 Mohamed Z. Hatim
            distance_matrix = 1 / (np.abs(associations.values) + 1e-10)
            np.fill_diagonal(distance_matrix, 0)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            dist_matrix = dijkstra(distance_matrix, directed=False)
            finite_distances = dist_matrix[np.isfinite(dist_matrix) & (dist_matrix > 0)]
            avg_path_length = np.mean(finite_distances) if len(finite_distances) > 0 else np.inf
        else:
            avg_path_length = np.nan
        
        return {
            'n_nodes': int(n_nodes),
            'n_edges': int(n_edges),
            'density': float(density),
            'mean_degree': float(np.mean(degrees)),
            'degree_distribution': degrees.tolist(),  # Convert to list to avoid pandas Series issues
            'avg_path_length': float(avg_path_length) if not np.isnan(avg_path_length) else np.nan
        }
    
    def _create_networkx_graph(self, associations: pd.DataFrame, 
                              threshold: float) -> 'nx.Graph':
        """Create NetworkX graph object."""
        if not NETWORKX_AVAILABLE:
            return None
        
        G = nx.Graph()
        
# Copyright (c) 2025 Mohamed Z. Hatim
        for species in associations.index:
            G.add_node(species)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        for i, species1 in enumerate(associations.index):
            for species2 in associations.index[i+1:]:
                association_strength = abs(associations.loc[species1, species2])
                if association_strength >= threshold:
                    G.add_edge(species1, species2, weight=association_strength)
        
        return G
    
    def _calculate_networkx_properties(self, G: 'nx.Graph') -> Dict[str, Any]:
        """Calculate advanced network properties using NetworkX."""
        if not NETWORKX_AVAILABLE or G is None:
            return {}
        
        props = {}
        
        try:
# Copyright (c) 2025 Mohamed Z. Hatim
            props['clustering_coefficient'] = nx.average_clustering(G)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            props['betweenness_centrality'] = nx.betweenness_centrality(G)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            try:
                props['eigenvector_centrality'] = nx.eigenvector_centrality(G, max_iter=1000)
            except:
                props['eigenvector_centrality'] = {}
            
# Copyright (c) 2025 Mohamed Z. Hatim
            props['n_components'] = nx.number_connected_components(G)
            props['largest_component_size'] = len(max(nx.connected_components(G), key=len))
            
# Copyright (c) 2025 Mohamed Z. Hatim
            if nx.is_connected(G):
                props['average_shortest_path'] = nx.average_shortest_path_length(G)
                
# Copyright (c) 2025 Mohamed Z. Hatim
                random_G = nx.erdos_renyi_graph(G.number_of_nodes(), 
                                              G.number_of_edges() / (G.number_of_nodes() * (G.number_of_nodes() - 1) / 2))
                if nx.is_connected(random_G):
                    props['small_world_sigma'] = (props['clustering_coefficient'] / nx.average_clustering(random_G)) / (props['average_shortest_path'] / nx.average_shortest_path_length(random_G))
            
        except Exception as e:
            warnings.warn(f"Some network properties could not be calculated: {str(e)}")
        
        return props


class CommunityAssemblyAnalyzer:
    """
    Analysis of community assembly processes and null models.
    """
    
    def __init__(self, random_state: int = 42):
        """Initialize CommunityAssemblyAnalyzer."""
        self.random_state = random_state
        np.random.seed(random_state)
        
    def assembly_process_analysis(self, 
                                community_data: pd.DataFrame,
                                trait_data: pd.DataFrame = None,
                                phylo_data: pd.DataFrame = None) -> Dict[str, Any]:
        """
        Analyze community assembly processes.
        
        Parameters
        ----------
        community_data : pd.DataFrame
            Community composition data (sites x species)
        trait_data : pd.DataFrame, optional
            Species trait data
        phylo_data : pd.DataFrame, optional
            Phylogenetic distance matrix
            
        Returns
        -------
        Dict[str, Any]
            Assembly process analysis results
        """
        results = {}
        
# Copyright (c) 2025 Mohamed Z. Hatim
        taxonomic_patterns = self._analyze_taxonomic_patterns(community_data)
        results['taxonomic'] = taxonomic_patterns
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if trait_data is not None:
            trait_patterns = self._analyze_trait_patterns(community_data, trait_data)
            results['trait_based'] = trait_patterns
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if phylo_data is not None:
            phylo_patterns = self._analyze_phylogenetic_patterns(community_data, phylo_data)
            results['phylogenetic'] = phylo_patterns
        
# Copyright (c) 2025 Mohamed Z. Hatim
        results['interpretation'] = self._interpret_assembly_processes(results)
        
        return results
    
    def _analyze_taxonomic_patterns(self, community_data: pd.DataFrame) -> Dict[str, Any]:
        """Analyze taxonomic diversity patterns."""
# Copyright (c) 2025 Mohamed Z. Hatim
        richness = community_data.sum(axis=1)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        evenness = []
        for site in community_data.index:
            site_data = community_data.loc[site]
            present_species = site_data[site_data > 0]
            if len(present_species) > 1:
# Copyright (c) 2025 Mohamed Z. Hatim
                proportions = present_species / present_species.sum()
                shannon = -np.sum(proportions * np.log(proportions))
                max_shannon = np.log(len(present_species))
                evenness.append(shannon / max_shannon if max_shannon > 0 else 0)
            else:
                evenness.append(0)
        
        return {
            'richness': richness,
            'evenness': pd.Series(evenness, index=community_data.index),
            'total_richness': (community_data.sum(axis=0) > 0).sum(),
            'mean_richness': richness.mean(),
            'richness_std': richness.std()
        }
    
    def _analyze_trait_patterns(self, community_data: pd.DataFrame, 
                               trait_data: pd.DataFrame) -> Dict[str, Any]:
        """Analyze trait-based assembly patterns."""
# Copyright (c) 2025 Mohamed Z. Hatim
        common_species = list(set(community_data.columns) & set(trait_data.index))
        if len(common_species) == 0:
            return {'error': 'No common species between community and trait data'}
        
        community_subset = community_data[common_species]
        trait_subset = trait_data.loc[common_species]
        
# Copyright (c) 2025 Mohamed Z. Hatim
        cwm_traits = []
        trait_variance = []
        
        for site in community_subset.index:
            site_abundances = community_subset.loc[site]
            present_species = site_abundances[site_abundances > 0].index
            
            if len(present_species) == 0:
                continue
            
            weights = site_abundances.loc[present_species]
            weights = weights / weights.sum()
            
            site_traits = trait_subset.loc[present_species]
            
# Copyright (c) 2025 Mohamed Z. Hatim
            cwm_site = {}
            variance_site = {}
            
            for trait in site_traits.select_dtypes(include=[np.number]).columns:
                trait_values = site_traits[trait].fillna(site_traits[trait].mean())
                cwm_site[trait] = (trait_values * weights).sum()
                
# Copyright (c) 2025 Mohamed Z. Hatim
                variance_site[trait] = ((trait_values - cwm_site[trait])**2 * weights).sum()
            
            cwm_traits.append(cwm_site)
            trait_variance.append(variance_site)
        
        cwm_df = pd.DataFrame(cwm_traits, index=community_subset.index[:len(cwm_traits)])
        variance_df = pd.DataFrame(trait_variance, index=community_subset.index[:len(trait_variance)])
        
        return {
            'cwm_traits': cwm_df,
            'trait_variance': variance_df,
            'species_used': common_species
        }
    
    def _analyze_phylogenetic_patterns(self, community_data: pd.DataFrame,
                                     phylo_data: pd.DataFrame) -> Dict[str, Any]:
        """Analyze phylogenetic assembly patterns."""
# Copyright (c) 2025 Mohamed Z. Hatim
        phylo_analyzer = PhylogeneticDiversityAnalyzer()
        phylo_analyzer.load_phylogeny(phylo_data)
        
        try:
            phylo_results = phylo_analyzer.calculate_phylogenetic_diversity(community_data)
            return phylo_results
        except Exception as e:
            return {'error': f'Phylogenetic analysis failed: {str(e)}'}
    
    def _interpret_assembly_processes(self, results: Dict[str, Any]) -> Dict[str, str]:
        """Interpret assembly processes based on patterns."""
        interpretation = {}
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if 'taxonomic' in results:
            richness_cv = results['taxonomic']['richness_std'] / results['taxonomic']['mean_richness']
            if richness_cv > 0.5:
                interpretation['richness_pattern'] = "High variation in species richness suggests environmental filtering"
            else:
                interpretation['richness_pattern'] = "Low variation in species richness suggests neutral assembly"
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if 'trait_based' in results and 'error' not in results['trait_based']:
            cwm_traits = results['trait_based']['cwm_traits']
            if not cwm_traits.empty:
                trait_cv = cwm_traits.std() / cwm_traits.mean()
                high_variation_traits = trait_cv[trait_cv > 0.3].index.tolist()
                if high_variation_traits:
                    interpretation['trait_pattern'] = f"High variation in {', '.join(high_variation_traits)} suggests environmental filtering on these traits"
                else:
                    interpretation['trait_pattern'] = "Low trait variation suggests weak environmental filtering"
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if 'phylogenetic' in results and 'metrics' in results['phylogenetic']:
            phylo_metrics = results['phylogenetic']['metrics']
            if not phylo_metrics.empty and 'nri' in phylo_metrics.columns:
                mean_nri = phylo_metrics['nri'].mean()
                if mean_nri > 1.96:  # Significant clustering
                    interpretation['phylogenetic_pattern'] = "Phylogenetic clustering suggests environmental filtering"
                elif mean_nri < -1.96:  # Significant overdispersion
                    interpretation['phylogenetic_pattern'] = "Phylogenetic overdispersion suggests competitive exclusion"
                else:
                    interpretation['phylogenetic_pattern'] = "Random phylogenetic pattern suggests neutral assembly"
        
        return interpretation


# Copyright (c) 2025 Mohamed Z. Hatim
def quick_phylogenetic_diversity(community_data: pd.DataFrame,
                                phylo_distances: pd.DataFrame) -> Dict[str, Any]:
    """
    Quick phylogenetic diversity analysis.
    
    Parameters
    ----------
    community_data : pd.DataFrame
        Community composition data
    phylo_distances : pd.DataFrame
        Phylogenetic distance matrix
        
    Returns
    -------
    Dict[str, Any]
        Phylogenetic diversity results
    """
    analyzer = PhylogeneticDiversityAnalyzer()
    analyzer.load_phylogeny(phylo_distances)
    return analyzer.calculate_phylogenetic_diversity(community_data)


def quick_metacommunity_analysis(community_data: pd.DataFrame) -> Dict[str, Any]:
    """
    Quick metacommunity structure analysis.
    
    Parameters
    ----------
    community_data : pd.DataFrame
        Community composition data
        
    Returns
    -------
    Dict[str, Any]
        Metacommunity analysis results
    """
    analyzer = MetacommunityAnalyzer()
    return analyzer.elements_of_metacommunity_structure(community_data)


def quick_cooccurrence_network(community_data: pd.DataFrame,
                              threshold: float = 0.5) -> Dict[str, Any]:
    """
    Quick species co-occurrence network analysis.
    
    Parameters
    ----------
    community_data : pd.DataFrame
        Community composition data
    threshold : float, optional
        Correlation threshold for network edges
        
    Returns
    -------
    Dict[str, Any]
        Network analysis results
    """
    analyzer = NetworkAnalyzer()
    return analyzer.build_cooccurrence_network(community_data, threshold=threshold)


#!/usr/bin/env python3
"""
VegZ Demo Script

This script demonstrates the main functionality of VegZ including
the comprehensive elbow analysis for optimal cluster determination.

Run with: VegZ-demo
"""

import pandas as pd
import numpy as np
from pathlib import Path

def main():
    """Main demo function."""
    print("=" * 60)
    print("VegZ Demonstration Script")
    print("Comprehensive Vegetation Data Analysis")
    print("=" * 60)
    
    try:
        from VegZ import VegZ, quick_elbow_analysis
        print("✓ VegZ imported successfully")
    except ImportError as e:
        print(f"✗ Could not import VegZ: {e}")
        return
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print("\n1. Creating synthetic vegetation data...")
    np.random.seed(42)
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    n_sites_per_type = 15
    n_species = 20
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    deciduous = np.random.poisson(5, (n_sites_per_type, n_species)) * \
                np.random.binomial(1, 0.4, (n_sites_per_type, n_species))
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    coniferous = np.random.poisson(8, (n_sites_per_type, n_species)) * \
                 np.random.binomial(1, 0.2, (n_sites_per_type, n_species))
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    mixed = np.random.poisson(6, (n_sites_per_type, n_species)) * \
            np.random.binomial(1, 0.3, (n_sites_per_type, n_species))
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    grassland = np.random.poisson(3, (n_sites_per_type, n_species)) * \
                np.random.binomial(1, 0.5, (n_sites_per_type, n_species))
    
    data = np.vstack([deciduous, coniferous, mixed, grassland])
    
    species_names = [f'Species_{i+1:02d}' for i in range(n_species)]
    site_names = [f'SITE_{i+1:03d}' for i in range(60)]
    
    vegetation_data = pd.DataFrame(data, index=site_names, columns=species_names)
    print(f"   Created data: {vegetation_data.shape[0]} sites × {vegetation_data.shape[1]} species")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print("\n2. Initializing VegZ...")
    veg = VegZ()
    veg.species_matrix = vegetation_data
    print("✓ VegZ initialized with synthetic data")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print("\n3. Calculating diversity indices...")
    diversity = veg.calculate_diversity(['shannon', 'simpson', 'richness', 'evenness'])
    print(f"   Calculated diversity for {len(diversity)} sites")
    print(f"   Mean Shannon diversity: {diversity['shannon'].mean():.2f}")
    print(f"   Mean species richness: {diversity['richness'].mean():.1f}")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print("\n4. Performing comprehensive elbow analysis...")
    elbow_results = veg.elbow_analysis(
        k_range=range(1, 12),
        methods=['knee_locator', 'derivative', 'variance_explained', 'distortion_jump'],
        plot_results=False  # Copyright (c) 2025 Mohamed Z. Hatim
    )
    
    print("   Elbow points detected by each method:")
    for method, k_value in elbow_results['elbow_points'].items():
        print(f"   • {method:20s}: k = {k_value}")
    
    optimal_k = elbow_results['recommendations']['consensus']
    confidence = elbow_results['recommendations']['confidence']
    print(f"\n   📊 Consensus recommendation: k = {optimal_k}")
    print(f"   📊 Confidence score: {confidence:.2f}")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print(f"\n5. Performing clustering with k = {optimal_k}...")
    clusters = veg.kmeans_clustering(n_clusters=optimal_k)
    print(f"   Clustering inertia: {clusters['inertia']:.1f}")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    cluster_counts = clusters['cluster_labels'].value_counts().sort_index()
    print("   Cluster sizes:")
    for cluster_id, count in cluster_counts.items():
        print(f"   • Cluster {cluster_id}: {count} sites")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print(f"\n6. Finding indicator species for {optimal_k} clusters...")
    indicators = veg.indicator_species_analysis(clusters['cluster_labels'])
    
    print("   Top indicator species per cluster:")
    for cluster_id in sorted(indicators['cluster'].unique()):
        cluster_indicators = indicators[indicators['cluster'] == cluster_id].nlargest(2, 'indicator_value')
        print(f"   • Cluster {cluster_id}:")
        for _, row in cluster_indicators.iterrows():
            print(f"     - {row['species']}: {row['indicator_value']:.1f}")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print("\n7. Performing ordination analysis...")
    pca_results = veg.pca_analysis(transform='hellinger', n_components=4)
    explained_var = pca_results['explained_variance_ratio']
    print(f"   PCA explained variance:")
    for i, var in enumerate(explained_var):
        print(f"   • PC{i+1}: {var:.1%}")
    print(f"   • Cumulative: {sum(explained_var):.1%}")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    nmds_results = veg.nmds_analysis(distance_metric='bray_curtis', n_dimensions=2)
    print(f"\n   NMDS stress value: {nmds_results['stress']:.3f}")
    if nmds_results['stress'] < 0.2:
        print("   ✓ Good NMDS representation (stress < 0.2)")
    else:
        print("   ⚠ High NMDS stress - consider more dimensions")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    print("\n8. Demonstrating quick functions...")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    quick_optimal_k = veg.quick_elbow_analysis(max_k=8)
    print(f"   Quick elbow analysis: k = {quick_optimal_k}")
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    stats = veg.summary_statistics()
    print(f"\n📈 Dataset Summary:")
    print(f"   • Total sites: {stats['n_sites']}")
    print(f"   • Total species: {stats['n_species']}")
    print(f"   • Mean species per site: {stats['mean_species_per_site']:.1f}")
    print(f"   • Total abundance: {stats['total_abundance']}")
    
    print("\n" + "=" * 60)
    print("✅ VegZ demonstration completed successfully!")
    print("📚 For more examples, see: https://vegz.readthedocs.io/")
    print("🐛 Report issues: https://github.com/mhatim99/VegZ/issues")
    print("=" * 60)


if __name__ == "__main__":
    main()


"""
Comprehensive temporal analysis module for vegetation and ecological data.

Copyright (c) 2025 Mohamed Z. Hatim
"""

import numpy as np
import pandas as pd
from typing import Union, List, Dict, Tuple, Optional, Any, Callable
from scipy import stats, optimize, signal
from scipy.interpolate import UnivariateSpline, interp1d
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures, StandardScaler
from sklearn.pipeline import Pipeline
from sklearn.metrics import r2_score, mean_squared_error
import warnings
from datetime import datetime, timedelta


class TemporalAnalyzer:
    """Comprehensive temporal analysis for ecological time series data."""
    
    def __init__(self):
        """Initialize temporal analyzer."""
        self.phenology_models = {
            'sigmoid': self._sigmoid_model,
            'double_sigmoid': self._double_sigmoid_model,
            'gaussian': self._gaussian_model,
            'beta': self._beta_model,
            'weibull': self._weibull_model
        }
        
        self.trend_methods = {
            'linear': self._linear_trend,
            'polynomial': self._polynomial_trend,
            'spline': self._spline_trend,
            'lowess': self._lowess_trend,
            'mann_kendall': self._mann_kendall_trend
        }
        
        self.decomposition_methods = {
            'classical': self._classical_decomposition,
            'stl': self._stl_decomposition,
            'x11': self._x11_decomposition
        }
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def phenology_modeling(self, data: pd.DataFrame,
                          time_col: str = 'date',
                          response_col: str = 'response',
                          model_type: str = 'sigmoid',
                          species_col: Optional[str] = None) -> Dict[str, Any]:
        """
        Model phenological patterns for species or communities.
        
        Parameters:
        -----------
        data : pd.DataFrame
            Time series data with date and response variables
        time_col : str
            Name of time/date column
        response_col : str
            Name of response variable (e.g., flowering, leaf emergence)
        model_type : str
            Type of phenological model to fit
        species_col : str, optional
            Column for grouping by species
            
        Returns:
        --------
        dict
            Phenology modeling results
        """
        if time_col not in data.columns or response_col not in data.columns:
            raise ValueError(f"Required columns {time_col} and {response_col} not found")
        
# Copyright (c) 2025 Mohamed Z. Hatim
        data_clean = data.copy()
        if not pd.api.types.is_numeric_dtype(data_clean[time_col]):
            dates = pd.to_datetime(data_clean[time_col])
            data_clean['day_of_year'] = dates.dt.dayofyear
            time_var = 'day_of_year'
        else:
            time_var = time_col
        
        results = {}
        
        if species_col and species_col in data.columns:
# Copyright (c) 2025 Mohamed Z. Hatim
            for species in data_clean[species_col].unique():
                species_data = data_clean[data_clean[species_col] == species]
                species_results = self._fit_phenology_model(
                    species_data[time_var].values,
                    species_data[response_col].values,
                    model_type
                )
                results[species] = species_results
        else:
# Copyright (c) 2025 Mohamed Z. Hatim
            results['combined'] = self._fit_phenology_model(
                data_clean[time_var].values,
                data_clean[response_col].values,
                model_type
            )
        
        return {
            'model_type': model_type,
            'results': results,
            'data_summary': {
                'n_observations': len(data_clean),
                'time_range': (data_clean[time_var].min(), data_clean[time_var].max()),
                'response_range': (data_clean[response_col].min(), data_clean[response_col].max())
            }
        }
    
    def _fit_phenology_model(self, x: np.ndarray, y: np.ndarray, 
                           model_type: str) -> Dict[str, Any]:
        """Fit individual phenology model."""
        if model_type not in self.phenology_models:
            raise ValueError(f"Unknown model type: {model_type}")
        
        model_func = self.phenology_models[model_type]
        
        try:
# Copyright (c) 2025 Mohamed Z. Hatim
            popt, pcov = optimize.curve_fit(
                model_func, x, y,
                maxfev=5000,
                bounds=self._get_model_bounds(model_type, x, y)
            )
            
# Copyright (c) 2025 Mohamed Z. Hatim
            y_pred = model_func(x, *popt)
            r_squared = r2_score(y, y_pred)
            rmse = np.sqrt(mean_squared_error(y, y_pred))
            
# Copyright (c) 2025 Mohamed Z. Hatim
            param_errors = np.sqrt(np.diag(pcov))
            
# Copyright (c) 2025 Mohamed Z. Hatim
            pheno_params = self._extract_phenology_parameters(model_type, popt, x)
            
            results = {
                'parameters': popt,
                'parameter_errors': param_errors,
                'covariance_matrix': pcov,
                'r_squared': r_squared,
                'rmse': rmse,
                'predicted_values': y_pred,
                'phenological_parameters': pheno_params,
                'model_function': model_func,
                'success': True
            }
            
        except Exception as e:
            warnings.warn(f"Model fitting failed: {e}")
            results = {
                'parameters': None,
                'error': str(e),
                'success': False
            }
        
        return results
    
    def _sigmoid_model(self, x: np.ndarray, a: float, b: float, 
                      c: float, d: float) -> np.ndarray:
        """Sigmoid (logistic) phenology model."""
        return a / (1 + np.exp(-b * (x - c))) + d
    
    def _double_sigmoid_model(self, x: np.ndarray, a1: float, b1: float, c1: float,
                             a2: float, b2: float, c2: float, d: float) -> np.ndarray:
        """Double sigmoid model for growing season."""
        rise = a1 / (1 + np.exp(-b1 * (x - c1)))
        fall = a2 / (1 + np.exp(b2 * (x - c2)))
        return rise - fall + d
    
    def _gaussian_model(self, x: np.ndarray, a: float, mu: float, 
                       sigma: float, d: float) -> np.ndarray:
        """Gaussian phenology model."""
        return a * np.exp(-0.5 * ((x - mu) / sigma)**2) + d
    
    def _beta_model(self, x: np.ndarray, a: float, alpha: float, 
                   beta: float, t1: float, t2: float) -> np.ndarray:
        """Beta function phenology model."""
# Copyright (c) 2025 Mohamed Z. Hatim
        x_norm = (x - t1) / (t2 - t1)
        x_norm = np.clip(x_norm, 0, 1)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        return a * (x_norm**(alpha-1)) * ((1-x_norm)**(beta-1))
    
    def _weibull_model(self, x: np.ndarray, a: float, k: float, 
                      lambda_: float, d: float) -> np.ndarray:
        """Weibull phenology model."""
        return a * (k / lambda_) * ((x / lambda_)**(k-1)) * np.exp(-((x / lambda_)**k)) + d
    
    def _get_model_bounds(self, model_type: str, x: np.ndarray, 
                         y: np.ndarray) -> Tuple[List, List]:
        """Get parameter bounds for optimization."""
        x_min, x_max = x.min(), x.max()
        y_min, y_max = y.min(), y.max()
        y_range = y_max - y_min
        
        if model_type == 'sigmoid':
            lower = [0, 0, x_min, y_min]
            upper = [y_range * 2, 1, x_max, y_max]
        elif model_type == 'double_sigmoid':
            lower = [0, 0, x_min, 0, 0, x_min, y_min]
            upper = [y_range * 2, 1, x_max, y_range * 2, 1, x_max, y_max]
        elif model_type == 'gaussian':
            lower = [0, x_min, 1, y_min]
            upper = [y_range * 2, x_max, (x_max - x_min), y_max]
        elif model_type == 'beta':
            lower = [0, 0.1, 0.1, x_min, x_min]
            upper = [y_range * 2, 10, 10, x_max, x_max]
        elif model_type == 'weibull':
            lower = [0, 0.1, 1, y_min]
            upper = [y_range * 2, 10, (x_max - x_min), y_max]
        else:
# Copyright (c) 2025 Mohamed Z. Hatim
            lower = [-np.inf] * 4
            upper = [np.inf] * 4
        
        return lower, upper
    
    def _extract_phenology_parameters(self, model_type: str, params: np.ndarray,
                                    x: np.ndarray) -> Dict[str, float]:
        """Extract meaningful phenological parameters."""
        pheno_params = {}
        
        if model_type == 'sigmoid':
            a, b, c, d = params
            pheno_params['onset'] = c
            pheno_params['amplitude'] = a
            pheno_params['rate'] = b
            pheno_params['baseline'] = d
            
        elif model_type == 'double_sigmoid':
            a1, b1, c1, a2, b2, c2, d = params
            pheno_params['green_up'] = c1
            pheno_params['senescence'] = c2
            pheno_params['peak_amplitude'] = a1 - a2 + d
            pheno_params['growing_season_length'] = c2 - c1
            
        elif model_type == 'gaussian':
            a, mu, sigma, d = params
            pheno_params['peak_time'] = mu
            pheno_params['peak_amplitude'] = a + d
            pheno_params['duration'] = 2 * sigma
            pheno_params['baseline'] = d
            
        elif model_type == 'beta':
            a, alpha, beta_param, t1, t2 = params
# Copyright (c) 2025 Mohamed Z. Hatim
            if alpha > 1 and beta_param > 1:
                peak_norm = (alpha - 1) / (alpha + beta_param - 2)
                pheno_params['peak_time'] = t1 + peak_norm * (t2 - t1)
            pheno_params['season_start'] = t1
            pheno_params['season_end'] = t2
            pheno_params['peak_amplitude'] = a
            
        return pheno_params
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def trend_detection(self, data: pd.DataFrame,
                       time_col: str = 'date',
                       response_col: str = 'response',
                       method: str = 'linear',
                       **kwargs) -> Dict[str, Any]:
        """
        Detect trends in ecological time series.
        
        Parameters:
        -----------
        data : pd.DataFrame
            Time series data
        time_col : str
            Time column name
        response_col : str
            Response variable column
        method : str
            Trend detection method
        **kwargs
            Additional parameters for specific methods
            
        Returns:
        --------
        dict
            Trend analysis results
        """
        if method not in self.trend_methods:
            raise ValueError(f"Unknown trend method: {method}")
        
# Copyright (c) 2025 Mohamed Z. Hatim
        data_clean = data.dropna(subset=[time_col, response_col])
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if not pd.api.types.is_numeric_dtype(data_clean[time_col]):
            time_numeric = pd.to_datetime(data_clean[time_col]).map(pd.Timestamp.toordinal)
        else:
            time_numeric = data_clean[time_col]
        
        x = time_numeric.values
        y = data_clean[response_col].values
        
# Copyright (c) 2025 Mohamed Z. Hatim
        trend_func = self.trend_methods[method]
        results = trend_func(x, y, **kwargs)
        
        results.update({
            'method': method,
            'data_summary': {
                'n_observations': len(data_clean),
                'time_span': x.max() - x.min(),
                'response_range': (y.min(), y.max())
            }
        })
        
        return results
    
    def _linear_trend(self, x: np.ndarray, y: np.ndarray, **kwargs) -> Dict[str, Any]:
        """Linear trend analysis."""
# Copyright (c) 2025 Mohamed Z. Hatim
        reg = LinearRegression()
        X = x.reshape(-1, 1)
        reg.fit(X, y)
        
        y_pred = reg.predict(X)
        slope = reg.coef_[0]
        intercept = reg.intercept_
        
# Copyright (c) 2025 Mohamed Z. Hatim
        r_squared = reg.score(X, y)
        n = len(y)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        residuals = y - y_pred
        mse = np.sum(residuals**2) / (n - 2)
        se_slope = np.sqrt(mse / np.sum((x - x.mean())**2))
        t_stat = slope / se_slope
        p_value = 2 * (1 - stats.t.cdf(abs(t_stat), n - 2))
        
        return {
            'slope': slope,
            'intercept': intercept,
            'r_squared': r_squared,
            'p_value': p_value,
            't_statistic': t_stat,
            'predicted_values': y_pred,
            'residuals': residuals,
            'trend_direction': 'increasing' if slope > 0 else 'decreasing' if slope < 0 else 'no trend',
            'trend_magnitude': abs(slope)
        }
    
    def _polynomial_trend(self, x: np.ndarray, y: np.ndarray, 
                         degree: int = 2, **kwargs) -> Dict[str, Any]:
        """Polynomial trend analysis."""
# Copyright (c) 2025 Mohamed Z. Hatim
        poly_features = PolynomialFeatures(degree=degree)
        X_poly = poly_features.fit_transform(x.reshape(-1, 1))
        
        reg = LinearRegression()
        reg.fit(X_poly, y)
        
        y_pred = reg.predict(X_poly)
        r_squared = reg.score(X_poly, y)
        
        return {
            'coefficients': reg.coef_,
            'intercept': reg.intercept_,
            'degree': degree,
            'r_squared': r_squared,
            'predicted_values': y_pred,
            'polynomial_features': poly_features
        }
    
    def _spline_trend(self, x: np.ndarray, y: np.ndarray, 
                     smoothing: float = None, **kwargs) -> Dict[str, Any]:
        """Spline trend analysis."""
# Copyright (c) 2025 Mohamed Z. Hatim
        if smoothing is None:
            smoothing = len(x)
        
        spline = UnivariateSpline(x, y, s=smoothing)
        y_pred = spline(x)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        x_fine = np.linspace(x.min(), x.max(), len(x) * 10)
        y_fine = spline(x_fine)
        derivatives = spline.derivative()(x_fine)
        
        return {
            'spline_object': spline,
            'predicted_values': y_pred,
            'smoothing_parameter': smoothing,
            'derivatives': derivatives,
            'fine_x': x_fine,
            'fine_y': y_fine,
            'mean_derivative': np.mean(derivatives),
            'trend_direction': 'increasing' if np.mean(derivatives) > 0 else 'decreasing' if np.mean(derivatives) < 0 else 'no trend'
        }
    
    def _lowess_trend(self, x: np.ndarray, y: np.ndarray, 
                     frac: float = 0.3, **kwargs) -> Dict[str, Any]:
        """LOWESS trend analysis."""
        try:
            from statsmodels.nonparametric.smoothers_lowess import lowess
            
# Copyright (c) 2025 Mohamed Z. Hatim
            smoothed = lowess(y, x, frac=frac, return_sorted=True)
            x_smooth = smoothed[:, 0]
            y_smooth = smoothed[:, 1]
            
# Copyright (c) 2025 Mohamed Z. Hatim
            interp_func = interp1d(x_smooth, y_smooth, 
                                 bounds_error=False, fill_value='extrapolate')
            y_pred = interp_func(x)
            
            return {
                'predicted_values': y_pred,
                'smooth_x': x_smooth,
                'smooth_y': y_smooth,
                'fraction': frac
            }
            
        except ImportError:
            warnings.warn("statsmodels not available, using simple moving average")
# Copyright (c) 2025 Mohamed Z. Hatim
            window = max(3, int(len(x) * frac))
            y_pred = pd.Series(y).rolling(window, center=True).mean().fillna(method='bfill').fillna(method='ffill').values
            
            return {
                'predicted_values': y_pred,
                'window_size': window,
                'method': 'moving_average_fallback'
            }
    
    def _mann_kendall_trend(self, x: np.ndarray, y: np.ndarray, 
                           alpha: float = 0.05, **kwargs) -> Dict[str, Any]:
        """Mann-Kendall trend test."""
        n = len(y)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        s = 0
        for i in range(n - 1):
            for j in range(i + 1, n):
                if y[j] > y[i]:
                    s += 1
                elif y[j] < y[i]:
                    s -= 1
        
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
        unique_vals, counts = np.unique(y, return_counts=True)
        tie_adjustment = np.sum(counts * (counts - 1) * (2 * counts + 5))
        
        var_s = (n * (n - 1) * (2 * n + 5) - tie_adjustment) / 18
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if s > 0:
            z = (s - 1) / np.sqrt(var_s)
        elif s < 0:
            z = (s + 1) / np.sqrt(var_s)
        else:
            z = 0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        p_value = 2 * (1 - stats.norm.cdf(abs(z)))
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if p_value < alpha:
            if s > 0:
                trend = 'increasing'
            else:
                trend = 'decreasing'
        else:
            trend = 'no significant trend'
        
# Copyright (c) 2025 Mohamed Z. Hatim
        slopes = []
        for i in range(n - 1):
            for j in range(i + 1, n):
                if x[j] != x[i]:
                    slopes.append((y[j] - y[i]) / (x[j] - x[i]))
        
        sens_slope = np.median(slopes) if slopes else 0
        
        return {
            's_statistic': s,
            'z_statistic': z,
            'p_value': p_value,
            'trend': trend,
            'sens_slope': sens_slope,
            'alpha': alpha,
            'significant': p_value < alpha
        }
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def seasonal_decomposition(self, data: pd.DataFrame,
                             time_col: str = 'date',
                             response_col: str = 'response',
                             method: str = 'classical',
                             period: Optional[int] = None,
                             **kwargs) -> Dict[str, Any]:
        """
        Decompose time series into trend, seasonal, and residual components.
        
        Parameters:
        -----------
        data : pd.DataFrame
            Time series data
        time_col : str
            Time column name
        response_col : str
            Response variable column
        method : str
            Decomposition method
        period : int, optional
            Seasonal period
        **kwargs
            Additional parameters
            
        Returns:
        --------
        dict
            Decomposition results
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        data_clean = data.dropna(subset=[time_col, response_col]).copy()
        data_clean[time_col] = pd.to_datetime(data_clean[time_col])
        data_clean = data_clean.sort_values(time_col)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        ts = pd.Series(
            data_clean[response_col].values,
            index=data_clean[time_col]
        )
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if period is None:
            period = self._detect_seasonality(ts)
        
        if method not in self.decomposition_methods:
            raise ValueError(f"Unknown decomposition method: {method}")
        
        decomp_func = self.decomposition_methods[method]
        return decomp_func(ts, period=period, **kwargs)
    
    def _classical_decomposition(self, ts: pd.Series, period: int,
                               model: str = 'additive', **kwargs) -> Dict[str, Any]:
        """Classical seasonal decomposition."""
        try:
            from statsmodels.tsa.seasonal import seasonal_decompose
            
# Copyright (c) 2025 Mohamed Z. Hatim
            decomposition = seasonal_decompose(ts, model=model, period=period)
            
            return {
                'trend': decomposition.trend,
                'seasonal': decomposition.seasonal,
                'residual': decomposition.resid,
                'observed': decomposition.observed,
                'model': model,
                'period': period,
                'method': 'classical'
            }
            
        except ImportError:
            warnings.warn("statsmodels not available, using simple decomposition")
# Copyright (c) 2025 Mohamed Z. Hatim
            return self._simple_decomposition(ts, period)
    
    def _stl_decomposition(self, ts: pd.Series, period: int, **kwargs) -> Dict[str, Any]:
        """STL (Seasonal and Trend decomposition using Loess) decomposition."""
        try:
            from statsmodels.tsa.seasonal import STL
            
# Copyright (c) 2025 Mohamed Z. Hatim
            stl = STL(ts, seasonal=kwargs.get('seasonal', 7), period=period)
            decomposition = stl.fit()
            
            return {
                'trend': decomposition.trend,
                'seasonal': decomposition.seasonal,
                'residual': decomposition.resid,
                'observed': decomposition.observed,
                'period': period,
                'method': 'stl'
            }
            
        except ImportError:
            warnings.warn("statsmodels not available, using simple decomposition")
            return self._simple_decomposition(ts, period)
    
    def _x11_decomposition(self, ts: pd.Series, period: int, **kwargs) -> Dict[str, Any]:
        """X-11 seasonal adjustment (simplified version)."""
# Copyright (c) 2025 Mohamed Z. Hatim
        return self._classical_decomposition(ts, period, model='multiplicative')
    
    def _simple_decomposition(self, ts: pd.Series, period: int) -> Dict[str, Any]:
        """Simple fallback decomposition method."""
# Copyright (c) 2025 Mohamed Z. Hatim
        trend = ts.rolling(window=period, center=True).mean()
        
# Copyright (c) 2025 Mohamed Z. Hatim
        detrended = ts - trend
        seasonal_means = detrended.groupby(detrended.index.dayofyear % period).mean()
        seasonal = detrended.copy()
        
        for i in range(len(seasonal)):
            day_of_cycle = seasonal.index[i].dayofyear % period
            if day_of_cycle in seasonal_means.index:
                seasonal.iloc[i] = seasonal_means[day_of_cycle]
            else:
                seasonal.iloc[i] = 0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        residual = ts - trend - seasonal
        
        return {
            'trend': trend,
            'seasonal': seasonal,
            'residual': residual,
            'observed': ts,
            'period': period,
            'method': 'simple'
        }
    
    def _detect_seasonality(self, ts: pd.Series) -> int:
        """Auto-detect seasonal period."""
# Copyright (c) 2025 Mohamed Z. Hatim
        if isinstance(ts.index, pd.DatetimeIndex):
            freq = pd.infer_freq(ts.index)
            if freq:
                if 'D' in freq:
                    return 365
                elif 'M' in freq:
                    return 12
                elif 'W' in freq:
                    return 52
                elif 'Q' in freq:
                    return 4
        
# Copyright (c) 2025 Mohamed Z. Hatim
        return max(4, min(len(ts) // 4, 365))
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def climate_vegetation_response(self, vegetation_data: pd.DataFrame,
                                  climate_data: pd.DataFrame,
                                  time_col: str = 'date',
                                  veg_col: str = 'response',
                                  climate_vars: List[str] = None,
                                  lag_periods: List[int] = [0, 1, 2, 3]) -> Dict[str, Any]:
        """
        Analyze vegetation response to climate variables.
        
        Parameters:
        -----------
        vegetation_data : pd.DataFrame
            Vegetation time series data
        climate_data : pd.DataFrame
            Climate data
        time_col : str
            Time column name
        veg_col : str
            Vegetation response column
        climate_vars : list
            Climate variables to analyze
        lag_periods : list
            Lag periods to test
            
        Returns:
        --------
        dict
            Climate-vegetation response analysis
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        veg_clean = vegetation_data.dropna(subset=[time_col, veg_col])
        climate_clean = climate_data.dropna(subset=[time_col])
        
# Copyright (c) 2025 Mohamed Z. Hatim
        merged = pd.merge(veg_clean, climate_clean, on=time_col, how='inner')
        
        if climate_vars is None:
# Copyright (c) 2025 Mohamed Z. Hatim
            climate_vars = merged.select_dtypes(include=[np.number]).columns.tolist()
            climate_vars = [col for col in climate_vars if col != veg_col]
        
        results = {}
        
        for climate_var in climate_vars:
            if climate_var not in merged.columns:
                continue
            
            var_results = {
                'correlations': {},
                'best_lag': None,
                'best_correlation': 0,
                'response_curve': None
            }
            
# Copyright (c) 2025 Mohamed Z. Hatim
            for lag in lag_periods:
                if lag == 0:
                    climate_lagged = merged[climate_var]
                    veg_response = merged[veg_col]
                else:
# Copyright (c) 2025 Mohamed Z. Hatim
                    climate_lagged = merged[climate_var].shift(lag)
                    veg_response = merged[veg_col]
                    
# Copyright (c) 2025 Mohamed Z. Hatim
                    valid_mask = ~(climate_lagged.isna() | veg_response.isna())
                    climate_lagged = climate_lagged[valid_mask]
                    veg_response = veg_response[valid_mask]
                
                if len(climate_lagged) < 3:
                    continue
                
# Copyright (c) 2025 Mohamed Z. Hatim
                correlation = np.corrcoef(climate_lagged, veg_response)[0, 1]
                if np.isnan(correlation):
                    correlation = 0
                
                var_results['correlations'][lag] = correlation
                
# Copyright (c) 2025 Mohamed Z. Hatim
                if abs(correlation) > abs(var_results['best_correlation']):
                    var_results['best_correlation'] = correlation
                    var_results['best_lag'] = lag
            
# Copyright (c) 2025 Mohamed Z. Hatim
            if var_results['best_lag'] is not None:
                best_lag = var_results['best_lag']
                
                if best_lag == 0:
                    x_vals = merged[climate_var].values
                    y_vals = merged[veg_col].values
                else:
                    climate_lagged = merged[climate_var].shift(best_lag)
                    valid_mask = ~(climate_lagged.isna() | merged[veg_col].isna())
                    x_vals = climate_lagged[valid_mask].values
                    y_vals = merged[veg_col][valid_mask].values
                
                if len(x_vals) > 3:
# Copyright (c) 2025 Mohamed Z. Hatim
                    response_curve = self._fit_response_curve(x_vals, y_vals)
                    var_results['response_curve'] = response_curve
            
            results[climate_var] = var_results
        
        return {
            'climate_variables': climate_vars,
            'lag_periods_tested': lag_periods,
            'results': results,
            'data_summary': {
                'n_observations': len(merged),
                'time_span': merged[time_col].max() - merged[time_col].min()
            }
        }
    
    def _fit_response_curve(self, x: np.ndarray, y: np.ndarray) -> Dict[str, Any]:
        """Fit response curve to climate-vegetation relationship."""
# Copyright (c) 2025 Mohamed Z. Hatim
        curve_types = {
            'linear': lambda x, a, b: a * x + b,
            'quadratic': lambda x, a, b, c: a * x**2 + b * x + c,
            'exponential': lambda x, a, b: a * np.exp(b * x),
            'logarithmic': lambda x, a, b: a * np.log(x) + b if np.all(x > 0) else a * x + b
        }
        
        best_curve = None
        best_r2 = -np.inf
        best_params = None
        
        for curve_name, curve_func in curve_types.items():
            try:
                if curve_name == 'logarithmic' and np.any(x <= 0):
                    continue
                    
                popt, _ = optimize.curve_fit(curve_func, x, y, maxfev=1000)
                y_pred = curve_func(x, *popt)
                r2 = r2_score(y, y_pred)
                
                if r2 > best_r2:
                    best_r2 = r2
                    best_curve = curve_name
                    best_params = popt
                    
            except:
                continue
        
        return {
            'best_curve_type': best_curve,
            'parameters': best_params,
            'r_squared': best_r2,
            'curve_function': curve_types[best_curve] if best_curve else None
        }
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def growth_curve_fitting(self, data: pd.DataFrame,
                           time_col: str = 'time',
                           size_col: str = 'size',
                           species_col: Optional[str] = None,
                           curve_type: str = 'logistic') -> Dict[str, Any]:
        """
        Fit growth curves to species data.
        
        Parameters:
        -----------
        data : pd.DataFrame
            Growth data
        time_col : str
            Time variable
        size_col : str
            Size/biomass variable
        species_col : str, optional
            Species grouping column
        curve_type : str
            Type of growth curve
            
        Returns:
        --------
        dict
            Growth curve fitting results
        """
        growth_models = {
            'logistic': self._logistic_growth,
            'gompertz': self._gompertz_growth,
            'von_bertalanffy': self._von_bertalanffy_growth,
            'exponential': self._exponential_growth,
            'power': self._power_growth
        }
        
        if curve_type not in growth_models:
            raise ValueError(f"Unknown curve type: {curve_type}")
        
        results = {}
        
        if species_col and species_col in data.columns:
# Copyright (c) 2025 Mohamed Z. Hatim
            for species in data[species_col].unique():
                species_data = data[data[species_col] == species]
                if len(species_data) < 4:  # Need minimum points for fitting
                    continue
                    
                species_results = self._fit_growth_curve(
                    species_data[time_col].values,
                    species_data[size_col].values,
                    growth_models[curve_type]
                )
                results[species] = species_results
        else:
# Copyright (c) 2025 Mohamed Z. Hatim
            results['combined'] = self._fit_growth_curve(
                data[time_col].values,
                data[size_col].values,
                growth_models[curve_type]
            )
        
        return {
            'curve_type': curve_type,
            'results': results,
            'model_function': growth_models[curve_type]
        }
    
    def _fit_growth_curve(self, t: np.ndarray, size: np.ndarray, 
                         model_func: Callable) -> Dict[str, Any]:
        """Fit individual growth curve."""
        try:
# Copyright (c) 2025 Mohamed Z. Hatim
            initial_guess = self._get_growth_initial_guess(t, size, model_func)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            popt, pcov = optimize.curve_fit(
                model_func, t, size,
                p0=initial_guess,
                maxfev=5000
            )
            
# Copyright (c) 2025 Mohamed Z. Hatim
            size_pred = model_func(t, *popt)
            r_squared = r2_score(size, size_pred)
            rmse = np.sqrt(mean_squared_error(size, size_pred))
            
# Copyright (c) 2025 Mohamed Z. Hatim
            growth_params = self._extract_growth_parameters(model_func, popt, t)
            
            return {
                'parameters': popt,
                'parameter_covariance': pcov,
                'r_squared': r_squared,
                'rmse': rmse,
                'predicted_values': size_pred,
                'growth_parameters': growth_params,
                'success': True
            }
            
        except Exception as e:
            return {
                'error': str(e),
                'success': False
            }
    
    def _logistic_growth(self, t: np.ndarray, K: float, r: float, t0: float) -> np.ndarray:
        """Logistic growth model."""
        return K / (1 + np.exp(-r * (t - t0)))
    
    def _gompertz_growth(self, t: np.ndarray, K: float, r: float, t0: float) -> np.ndarray:
        """Gompertz growth model."""
        return K * np.exp(-np.exp(-r * (t - t0)))
    
    def _von_bertalanffy_growth(self, t: np.ndarray, Linf: float, K: float, t0: float) -> np.ndarray:
        """von Bertalanffy growth model."""
        return Linf * (1 - np.exp(-K * (t - t0)))**3
    
    def _exponential_growth(self, t: np.ndarray, N0: float, r: float) -> np.ndarray:
        """Exponential growth model."""
        return N0 * np.exp(r * t)
    
    def _power_growth(self, t: np.ndarray, a: float, b: float) -> np.ndarray:
        """Power growth model."""
        return a * (t ** b)
    
    def _get_growth_initial_guess(self, t: np.ndarray, size: np.ndarray, 
                                 model_func: Callable) -> List[float]:
        """Get initial parameter guesses for growth curves."""
        t_range = t.max() - t.min()
        size_max = size.max()
        size_min = size.min()
        
        if model_func == self._logistic_growth:
            return [size_max * 1.2, 0.1, t.mean()]
        elif model_func == self._gompertz_growth:
            return [size_max * 1.2, 0.1, t.mean()]
        elif model_func == self._von_bertalanffy_growth:
            return [size_max * 1.2, 0.1, t.min()]
        elif model_func == self._exponential_growth:
            return [size_min, 0.1]
        elif model_func == self._power_growth:
            return [1.0, 1.0]
        else:
            return [1.0] * 3  # Default
    
    def _extract_growth_parameters(self, model_func: Callable, params: np.ndarray,
                                 t: np.ndarray) -> Dict[str, float]:
        """Extract meaningful growth parameters."""
        growth_params = {}
        
        if model_func == self._logistic_growth:
            K, r, t0 = params
            growth_params['carrying_capacity'] = K
            growth_params['growth_rate'] = r
            growth_params['inflection_point'] = t0
            growth_params['max_growth_rate_time'] = t0
            
        elif model_func == self._gompertz_growth:
            K, r, t0 = params
            growth_params['asymptotic_size'] = K
            growth_params['growth_rate'] = r
            growth_params['inflection_point'] = t0
            
        elif model_func == self._exponential_growth:
            N0, r = params
            growth_params['initial_size'] = N0
            growth_params['growth_rate'] = r
            
        return growth_params


"""
Tests for VegZ core functionality.
"""

import pytest
import pandas as pd
import numpy as np
from pathlib import Path

from VegZ import VegZ, quick_diversity_analysis, quick_elbow_analysis


class TestVegZCore:
    """Test core VegZ functionality."""
    
    @pytest.fixture
    def sample_data(self):
        """Create sample vegetation data for testing."""
# Copyright (c) 2025 Mohamed Z. Hatim
        np.random.seed(42)
        n_sites = 30
        n_species = 15
        
# Copyright (c) 2025 Mohamed Z. Hatim
        group1 = np.random.poisson(5, (8, n_species)) * np.random.binomial(1, 0.4, (8, n_species))
# Copyright (c) 2025 Mohamed Z. Hatim
        group2 = np.random.poisson(3, (7, n_species)) * np.random.binomial(1, 0.3, (7, n_species))
# Copyright (c) 2025 Mohamed Z. Hatim
        group3 = np.random.poisson(8, (8, n_species)) * np.random.binomial(1, 0.2, (8, n_species))
# Copyright (c) 2025 Mohamed Z. Hatim
        group4 = np.random.poisson(2, (7, n_species)) * np.random.binomial(1, 0.5, (7, n_species))
        
        data = np.vstack([group1, group2, group3, group4])
        
        species_names = [f'Species_{i+1:02d}' for i in range(n_species)]
        site_names = [f'SITE_{i+1:03d}' for i in range(n_sites)]
        
        return pd.DataFrame(data, index=site_names, columns=species_names)
    
    def test_vegz_initialization(self):
        """Test VegZ initialization."""
        veg = VegZ()
        assert veg.data is None
        assert veg.species_matrix is None
        assert veg.environmental_data is None
        assert isinstance(veg.metadata, dict)
    
    def test_species_matrix_assignment(self, sample_data):
        """Test species matrix assignment."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        assert veg.species_matrix is not None
        assert veg.species_matrix.shape == sample_data.shape
        assert list(veg.species_matrix.columns) == list(sample_data.columns)
    
    def test_diversity_calculation(self, sample_data):
        """Test diversity indices calculation."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        diversity = veg.calculate_diversity(['shannon', 'simpson', 'richness'])
        
        assert isinstance(diversity, pd.DataFrame)
        assert 'shannon' in diversity.columns
        assert 'simpson' in diversity.columns
        assert 'richness' in diversity.columns
        assert len(diversity) == len(sample_data)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        assert all(diversity['shannon'] >= 0)
        assert all(diversity['simpson'] >= 0)
        assert all(diversity['simpson'] <= 1)
        assert all(diversity['richness'] >= 0)
    
    def test_pca_analysis(self, sample_data):
        """Test PCA analysis."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        pca_results = veg.pca_analysis(transform='hellinger', n_components=4)
        
        assert isinstance(pca_results, dict)
        assert 'scores' in pca_results
        assert 'loadings' in pca_results
        assert 'explained_variance_ratio' in pca_results
        
        assert pca_results['scores'].shape[0] == len(sample_data)
        assert pca_results['scores'].shape[1] == 4
        assert len(pca_results['explained_variance_ratio']) == 4
    
    def test_nmds_analysis(self, sample_data):
        """Test NMDS analysis."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        nmds_results = veg.nmds_analysis(distance_metric='bray_curtis', n_dimensions=2)
        
        assert isinstance(nmds_results, dict)
        assert 'scores' in nmds_results
        assert 'stress' in nmds_results
        
        assert nmds_results['scores'].shape[0] == len(sample_data)
        assert nmds_results['scores'].shape[1] == 2
        assert isinstance(nmds_results['stress'], float)
    
    def test_hierarchical_clustering(self, sample_data):
        """Test hierarchical clustering."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        hier_results = veg.hierarchical_clustering(n_clusters=4)
        
        assert isinstance(hier_results, dict)
        assert 'linkage_matrix' in hier_results
        assert 'cluster_labels' in hier_results
        
        assert len(hier_results['cluster_labels']) == len(sample_data)
        assert hier_results['cluster_labels'].max() <= 4
    
    def test_kmeans_clustering(self, sample_data):
        """Test k-means clustering."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        kmeans_results = veg.kmeans_clustering(n_clusters=3)
        
        assert isinstance(kmeans_results, dict)
        assert 'cluster_labels' in kmeans_results
        assert 'centroids' in kmeans_results
        assert 'inertia' in kmeans_results
        
        assert len(kmeans_results['cluster_labels']) == len(sample_data)
        assert kmeans_results['cluster_labels'].max() < 3  # 0-indexed
    
    def test_elbow_analysis(self, sample_data):
        """Test elbow analysis functionality."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        elbow_results = veg.elbow_analysis(
            k_range=range(1, 8),
            methods=['knee_locator', 'derivative'],
            plot_results=False
        )
        
        assert isinstance(elbow_results, dict)
        assert 'elbow_points' in elbow_results
        assert 'recommendations' in elbow_results
        assert 'metrics' in elbow_results
        
# Copyright (c) 2025 Mohamed Z. Hatim
        assert 'knee_locator' in elbow_results['elbow_points']
        assert 'derivative' in elbow_results['elbow_points']
        
# Copyright (c) 2025 Mohamed Z. Hatim
        assert 'consensus' in elbow_results['recommendations']
    
    def test_quick_elbow_analysis(self, sample_data):
        """Test quick elbow analysis."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
        optimal_k = veg.quick_elbow_analysis(max_k=8)
        
        assert isinstance(optimal_k, int)
        assert 1 <= optimal_k <= 8
    
    def test_indicator_species_analysis(self, sample_data):
        """Test indicator species analysis."""
        veg = VegZ()
        veg.species_matrix = sample_data
        
# Copyright (c) 2025 Mohamed Z. Hatim
        clusters = veg.kmeans_clustering(n_clusters=3)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        indicators = veg.indicator_species_analysis(clusters['cluster_labels'])
        
        assert isinstance(indicators, pd.DataFrame)
        assert 'species' in indicators.columns
        assert 'cluster' in indicators.columns
        assert 'indicator_value' in indicators.columns
        
# Copyright (c) 2025 Mohamed Z. Hatim
        assert all(indicators['indicator_value'] >= 0)
        assert all(indicators['indicator_value'] <= 100)


class TestQuickFunctions:
    """Test quick analysis functions."""
    
    @pytest.fixture
    def sample_data(self):
        """Create sample data for testing."""
        np.random.seed(42)
        data = np.random.poisson(3, (20, 10)) * np.random.binomial(1, 0.3, (20, 10))
        
        species_names = [f'Sp_{i+1}' for i in range(10)]
        site_names = [f'Site_{i+1}' for i in range(20)]
        
        return pd.DataFrame(data, index=site_names, columns=species_names)
    
    def test_quick_diversity_analysis(self, sample_data):
        """Test quick diversity analysis function."""
        diversity = quick_diversity_analysis(sample_data)
        
        assert isinstance(diversity, pd.DataFrame)
        assert len(diversity) == len(sample_data)
        assert 'shannon' in diversity.columns
        assert 'simpson' in diversity.columns
        assert 'richness' in diversity.columns
    
    def test_quick_elbow_analysis(self, sample_data):
        """Test quick elbow analysis function."""
        results = quick_elbow_analysis(sample_data, max_k=6, plot_results=False)
        
        assert isinstance(results, dict)
        assert 'recommendations' in results
        assert 'elbow_points' in results
        assert results['recommendations']['consensus'] is not None


class TestDataTransformations:
    """Test data transformation methods."""
    
    @pytest.fixture
    def sample_data(self):
        """Create sample data with known properties."""
        data = np.array([
            [10, 5, 0, 2],
            [0, 8, 3, 1],
            [5, 0, 7, 4]
        ])
        return pd.DataFrame(data, columns=['A', 'B', 'C', 'D'])
    
    def test_hellinger_transformation(self, sample_data):
        """Test Hellinger transformation."""
        veg = VegZ()
        transformed = veg._transform_data(sample_data, 'hellinger')
        
        assert transformed.shape == sample_data.shape
# Copyright (c) 2025 Mohamed Z. Hatim
        assert np.all(transformed >= 0)
        assert np.all(transformed <= 1)
    
    def test_log_transformation(self, sample_data):
        """Test log transformation."""
        veg = VegZ()
        transformed = veg._transform_data(sample_data, 'log')
        
        assert transformed.shape == sample_data.shape
# Copyright (c) 2025 Mohamed Z. Hatim
        assert np.all(np.isfinite(transformed))
    
    def test_sqrt_transformation(self, sample_data):
        """Test sqrt transformation."""
        veg = VegZ()
        transformed = veg._transform_data(sample_data, 'sqrt')
        
        assert transformed.shape == sample_data.shape
        assert np.all(transformed >= 0)  # Square root is always non-negative


if __name__ == '__main__':
    pytest.main([__file__])


"""
Comprehensive multivariate analysis module for vegetation data.

Copyright (c) 2025 Mohamed Z. Hatim
"""

import numpy as np
import pandas as pd
from typing import Union, List, Dict, Tuple, Optional, Any
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.linalg import svd, eig
from sklearn.decomposition import PCA
from sklearn.manifold import MDS
from sklearn.preprocessing import StandardScaler
import warnings


class MultivariateAnalyzer:
    """Comprehensive multivariate analysis for ecological communities."""
    
    def __init__(self):
        """Initialize multivariate analyzer."""
        self.available_methods = [
            'pca', 'ca', 'dca', 'cca', 'rda', 'nmds', 'pcoa'
        ]
        
        self.distance_metrics = {
            'bray_curtis': self._bray_curtis_distance,
            'jaccard': self._jaccard_distance,
            'sorensen': self._sorensen_distance,
            'euclidean': self._euclidean_distance,
            'manhattan': self._manhattan_distance,
            'canberra': self._canberra_distance,
            'chord': self._chord_distance,
            'hellinger': self._hellinger_distance
        }

        # Copyright (c) 2025 Mohamed Z. Hatim
        # Copyright (c) 2025 Mohamed Z. Hatim
        # Copyright (c) 2025 Mohamed Z. Hatim

    def pca_analysis(self, data: pd.DataFrame,
                    transform: str = 'hellinger',
                    n_components: Optional[int] = None) -> Dict[str, Any]:
        """
        Principal Component Analysis.
        
        Parameters
        ----------
        data : pd.DataFrame
            Community composition data (sites x species)
        transform : str, optional
            Data transformation method, by default 'hellinger'
        n_components : int, optional
            Number of components to retain, by default None (all)
            
        Returns
        -------
        Dict[str, Any]
            PCA results including site scores, species scores, and explained variance
        """
        from sklearn.decomposition import PCA
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        if transform == 'hellinger':
    # Copyright (c) 2025 Mohamed Z. Hatim
            transformed_data = np.sqrt(data.div(data.sum(axis=1), axis=0).fillna(0))
        elif transform == 'log':
            transformed_data = np.log1p(data)
        elif transform == 'sqrt':
            transformed_data = np.sqrt(data)
        elif transform == 'standardize':
            scaler = StandardScaler()
            transformed_data = pd.DataFrame(
                scaler.fit_transform(data),
                index=data.index,
                columns=data.columns
            )
        else:
            transformed_data = data
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        if n_components is None:
            n_components = min(data.shape[0] - 1, data.shape[1])
        
        pca = PCA(n_components=n_components)
        site_scores = pca.fit_transform(transformed_data)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        species_scores = pca.components_.T * np.sqrt(pca.explained_variance_)
        
        results = {
            'method': 'PCA',
            'site_scores': pd.DataFrame(
                site_scores,
                index=data.index,
                columns=[f'PC{i+1}' for i in range(site_scores.shape[1])]
            ),
            'species_scores': pd.DataFrame(
                species_scores,
                index=data.columns,
                columns=[f'PC{i+1}' for i in range(species_scores.shape[1])]
            ),
            'explained_variance_ratio': pca.explained_variance_ratio_,
            'explained_variance': pca.explained_variance_,
            'eigenvalues': pca.explained_variance_,
            'total_variance': np.sum(pca.explained_variance_),
            'transform': transform,
            'pca_object': pca
        }
        
        return results
    
    def nmds_analysis(self, data: pd.DataFrame,
                     distance_metric: str = 'bray_curtis',
                     n_dimensions: int = 2,
                     max_iterations: int = 300,
                     random_state: int = 42) -> Dict[str, Any]:
        """
        Non-metric Multidimensional Scaling (NMDS).
        
        Parameters
        ----------
        data : pd.DataFrame
            Community composition data (sites x species)
        distance_metric : str, optional
            Distance metric to use, by default 'bray_curtis'
        n_dimensions : int, optional
            Number of dimensions, by default 2
        max_iterations : int, optional
            Maximum number of iterations, by default 300
        random_state : int, optional
            Random state for reproducibility, by default 42
            
        Returns
        -------
        Dict[str, Any]
            NMDS results including site scores and stress value
        """
        from sklearn.manifold import MDS
        from scipy.spatial.distance import squareform, pdist
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        if distance_metric == 'bray_curtis':
            distances = pdist(data.values, metric='braycurtis')
        elif distance_metric == 'jaccard':
            distances = pdist(data.values, metric='jaccard')
        elif distance_metric == 'euclidean':
            distances = pdist(data.values, metric='euclidean')
        else:
            distances = pdist(data.values, metric=distance_metric)
        
        distance_matrix = squareform(distances)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        mds = MDS(
            n_components=n_dimensions,
            metric=False,  # Non-metric MDS
            dissimilarity='precomputed',
            max_iter=max_iterations,
            random_state=random_state
        )
        
        site_scores = mds.fit_transform(distance_matrix)
        
        results = {
            'method': 'NMDS',
            'site_scores': pd.DataFrame(
                site_scores,
                index=data.index,
                columns=[f'NMDS{i+1}' for i in range(n_dimensions)]
            ),
            'stress': mds.stress_,
            'n_iterations': mds.n_iter_,
            'distance_metric': distance_metric,
            'distance_matrix': pd.DataFrame(
                distance_matrix,
                index=data.index,
                columns=data.index
            ),
            'mds_object': mds
        }
        
        return results

    def ca_analysis(self, data: pd.DataFrame,
                   scaling: int = 1) -> Dict[str, Any]:
        """
        Correspondence Analysis (CA) - reciprocal averaging.

        Alias for correspondence_analysis(). Use this abbreviated form for new code.

        Parameters:
        -----------
        data : pd.DataFrame
            Species abundance matrix (sites x species)
        scaling : int
            Scaling type (1 or 2)

        Returns:
        --------
        dict
            CA results including scores, eigenvalues, and diagnostics
        """
        return self.correspondence_analysis(data, scaling)

    def correspondence_analysis(self, data: pd.DataFrame, scaling: int = 1) -> Dict[str, Any]:
        """
        Correspondence Analysis (CA).

        This method is kept for backward compatibility.
        Use ca_analysis() for new code.
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        data_clean = data.loc[(data.sum(axis=1) > 0), (data.sum(axis=0) > 0)]

        if data_clean.empty:
            raise ValueError("Data matrix is empty after removing zero rows/columns")

    # Copyright (c) 2025 Mohamed Z. Hatim
        X = data_clean.values.astype(float)

    # Copyright (c) 2025 Mohamed Z. Hatim
        n, p = X.shape

    # Copyright (c) 2025 Mohamed Z. Hatim
        grand_total = X.sum()

    # Copyright (c) 2025 Mohamed Z. Hatim
        row_totals = X.sum(axis=1)
        col_totals = X.sum(axis=0)

    # Copyright (c) 2025 Mohamed Z. Hatim
        expected = np.outer(row_totals, col_totals) / grand_total

    # Copyright (c) 2025 Mohamed Z. Hatim
        residuals = (X - expected) / np.sqrt(expected)

    # Copyright (c) 2025 Mohamed Z. Hatim
        residuals = np.nan_to_num(residuals, nan=0.0, posinf=0.0, neginf=0.0)

    # Copyright (c) 2025 Mohamed Z. Hatim
        U, s, Vt = svd(residuals, full_matrices=False)
        V = Vt.T

    # Copyright (c) 2025 Mohamed Z. Hatim
        eigenvalues = s**2

    # Copyright (c) 2025 Mohamed Z. Hatim
        total_inertia = eigenvalues.sum()

    # Copyright (c) 2025 Mohamed Z. Hatim
        explained_variance = eigenvalues / total_inertia if total_inertia > 0 else eigenvalues

    # Copyright (c) 2025 Mohamed Z. Hatim
        if scaling == 1:
            site_scores = U @ np.diag(s) / np.sqrt(row_totals[:, np.newaxis])
            species_scores = V
        else:  # scaling 2: distances among species preserved
            site_scores = U
            species_scores = V @ np.diag(s) / np.sqrt(col_totals[:, np.newaxis])

    # Copyright (c) 2025 Mohamed Z. Hatim
        site_scores = np.nan_to_num(site_scores, nan=0.0, posinf=0.0, neginf=0.0)
        species_scores = np.nan_to_num(species_scores, nan=0.0, posinf=0.0, neginf=0.0)

        results = {
            'site_scores': pd.DataFrame(
                site_scores,
                index=data_clean.index,
                columns=[f'CA{i+1}' for i in range(site_scores.shape[1])]
            ),
            'species_scores': pd.DataFrame(
                species_scores,
                index=data_clean.columns,
                columns=[f'CA{i+1}' for i in range(species_scores.shape[1])]
            ),
            'eigenvalues': eigenvalues,
            'explained_variance_ratio': explained_variance,
            'total_inertia': total_inertia,
            'scaling': scaling,
            'method': 'CA'
        }

        return results

    def dca_analysis(self, data: pd.DataFrame,
                     segments: int = 26) -> Dict[str, Any]:
        """
        Detrended Correspondence Analysis (DCA).
        
        Parameters:
        -----------
        data : pd.DataFrame
            Species abundance matrix
        segments : int
            Number of segments for detrending
            
        Returns:
        --------
        dict
            DCA results
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        ca_results = self.ca_analysis(data)
        
        site_scores = ca_results['site_scores'].values
        species_scores = ca_results['species_scores'].values
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        detrended_site_scores = site_scores.copy()
        detrended_species_scores = species_scores.copy()
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        if site_scores.shape[1] > 1:
    # Copyright (c) 2025 Mohamed Z. Hatim
            axis1_range = site_scores[:, 0].max() - site_scores[:, 0].min()
            if axis1_range > 0:
                segment_width = axis1_range / segments
            else:
                segment_width = 1.0
            
            for i in range(segments):
                segment_min = site_scores[:, 0].min() + i * segment_width
                segment_max = segment_min + segment_width
                
    # Copyright (c) 2025 Mohamed Z. Hatim
                in_segment = (site_scores[:, 0] >= segment_min) & (site_scores[:, 0] <= segment_max)
                
                if in_segment.sum() > 1:
    # Copyright (c) 2025 Mohamed Z. Hatim
                    segment_mean = site_scores[in_segment, 1].mean()
                    
    # Copyright (c) 2025 Mohamed Z. Hatim
                    detrended_site_scores[in_segment, 1] -= segment_mean
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        for axis in range(detrended_site_scores.shape[1]):
            if detrended_site_scores[:, axis].std() > 0:
                detrended_site_scores[:, axis] /= detrended_site_scores[:, axis].std()
        
        results = {
            'site_scores': pd.DataFrame(
                detrended_site_scores,
                index=ca_results['site_scores'].index,
                columns=[f'DCA{i+1}' for i in range(detrended_site_scores.shape[1])]
            ),
            'species_scores': pd.DataFrame(
                detrended_species_scores,
                index=ca_results['species_scores'].index,
                columns=[f'DCA{i+1}' for i in range(detrended_species_scores.shape[1])]
            ),
            'eigenvalues': ca_results['eigenvalues'],
            'gradient_lengths': self._calculate_gradient_lengths(detrended_site_scores),
            'method': 'DCA'
        }

        return results

    def detrended_correspondence_analysis(self, data: pd.DataFrame, segments: int = 26) -> Dict[str, Any]:
        """
        Detrended Correspondence Analysis (DCA) - alias for dca_analysis.

        This method is kept for backward compatibility.
        Use dca_analysis() for new code.
        """
        return self.dca_analysis(data, segments)

    def canonical_correspondence_analysis(self, species_data: pd.DataFrame,
                                        env_data: pd.DataFrame,
                                        scaling: int = 1) -> Dict[str, Any]:
        """
        Canonical Correspondence Analysis (CCA).
        
        Parameters:
        -----------
        species_data : pd.DataFrame
            Species abundance matrix
        env_data : pd.DataFrame
            Environmental variables matrix
        scaling : int
            Scaling type
            
        Returns:
        --------
        dict
            CCA results
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        common_index = species_data.index.intersection(env_data.index)
        species_aligned = species_data.loc[common_index]
        env_aligned = env_data.loc[common_index]
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        complete_cases = ~env_aligned.isnull().any(axis=1)
        species_complete = species_aligned.loc[complete_cases]
        env_complete = env_aligned.loc[complete_cases]
        
        if len(species_complete) == 0:
            raise ValueError("No complete cases found")
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        scaler = StandardScaler()
        env_std = scaler.fit_transform(env_complete)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        X = species_complete.values
        Z = env_std
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        row_totals = X.sum(axis=1)
        row_weights = row_totals / row_totals.sum()
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        W_diag = np.diag(row_weights)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        Y = X / row_totals[:, np.newaxis]  # Profiles
        Y = np.nan_to_num(Y, nan=0.0, posinf=0.0, neginf=0.0)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        try:
    # Copyright (c) 2025 Mohamed Z. Hatim
            U, s, Vt = svd(Z.T @ W_diag @ Y, full_matrices=False)
            
            eigenvalues = s**2
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            site_scores = Z @ U
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            species_scores = Y.T @ W_diag @ Z @ U
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            env_scores = U.T
            
        except np.linalg.LinAlgError:
            warnings.warn("CCA computation failed, using simplified approach")
    # Copyright (c) 2025 Mohamed Z. Hatim
            eigenvalues = np.array([1.0])
            site_scores = np.column_stack([np.arange(len(species_complete))])
            species_scores = np.column_stack([np.arange(len(species_complete.columns))])
            env_scores = np.column_stack([np.arange(len(env_complete.columns))])
        
        results = {
            'site_scores': pd.DataFrame(
                site_scores,
                index=species_complete.index,
                columns=[f'CCA{i+1}' for i in range(site_scores.shape[1])]
            ),
            'species_scores': pd.DataFrame(
                species_scores,
                index=species_complete.columns,
                columns=[f'CCA{i+1}' for i in range(species_scores.shape[1])]
            ),
            'env_scores': pd.DataFrame(
                env_scores.T,
                index=env_complete.columns,
                columns=[f'CCA{i+1}' for i in range(env_scores.shape[0])]
            ),
            'eigenvalues': eigenvalues,
            'explained_variance_ratio': eigenvalues / eigenvalues.sum() if eigenvalues.sum() > 0 else eigenvalues,
            'scaling': scaling,
            'method': 'CCA'
        }
        
        return results
    
    def cca_analysis(self, species_data: pd.DataFrame, env_data: pd.DataFrame,
                    scaling: int = 1) -> Dict[str, Any]:
        """
        Canonical Correspondence Analysis (CCA) - alias for canonical_correspondence_analysis.
        
        Parameters:
        -----------
        species_data : pd.DataFrame
            Species abundance matrix
        env_data : pd.DataFrame
            Environmental variables matrix
        scaling : int
            Scaling type
            
        Returns:
        --------
        dict
            CCA results
        """
        return self.canonical_correspondence_analysis(species_data, env_data, scaling)
    
    def rda_analysis(self, species_data: pd.DataFrame,
                     env_data: pd.DataFrame) -> Dict[str, Any]:
        """
        Redundancy Analysis (RDA).
        
        Parameters:
        -----------
        species_data : pd.DataFrame
            Species abundance matrix
        env_data : pd.DataFrame
            Environmental variables matrix
            
        Returns:
        --------
        dict
            RDA results
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        common_index = species_data.index.intersection(env_data.index)
        species_aligned = species_data.loc[common_index]
        env_aligned = env_data.loc[common_index]
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        complete_cases = ~env_aligned.isnull().any(axis=1)
        species_complete = species_aligned.loc[complete_cases]
        env_complete = env_aligned.loc[complete_cases]
        
        if len(species_complete) == 0:
            raise ValueError("No complete cases found")
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        species_scaler = StandardScaler()
        env_scaler = StandardScaler()
        
        Y = species_scaler.fit_transform(species_complete)
        X = env_scaler.fit_transform(env_complete)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        try:
    # Copyright (c) 2025 Mohamed Z. Hatim
    # Copyright (c) 2025 Mohamed Z. Hatim
            XtX_inv = np.linalg.pinv(X.T @ X)
            projection_matrix = X @ XtX_inv @ X.T
            Y_fitted = projection_matrix @ Y
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            U, s, Vt = svd(Y_fitted, full_matrices=False)
            
            eigenvalues = s**2
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            site_scores_constrained = U @ np.diag(s)
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            species_scores = Vt.T
            
    # Copyright (c) 2025 Mohamed Z. Hatim
            env_scores = XtX_inv @ X.T @ Y_fitted
            
        except np.linalg.LinAlgError:
            warnings.warn("RDA computation failed, using PCA fallback")
            pca = PCA()
            site_scores_constrained = pca.fit_transform(Y)
            species_scores = pca.components_.T
            eigenvalues = pca.explained_variance_
            env_scores = np.random.randn(len(env_complete.columns), len(eigenvalues))
        
        results = {
            'site_scores': pd.DataFrame(
                site_scores_constrained,
                index=species_complete.index,
                columns=[f'RDA{i+1}' for i in range(site_scores_constrained.shape[1])]
            ),
            'species_scores': pd.DataFrame(
                species_scores,
                index=species_complete.columns,
                columns=[f'RDA{i+1}' for i in range(species_scores.shape[1])]
            ),
            'env_scores': pd.DataFrame(
                env_scores,
                index=env_complete.columns,
                columns=[f'RDA{i+1}' for i in range(env_scores.shape[1])]
            ),
            'eigenvalues': eigenvalues,
            'explained_variance_ratio': eigenvalues / eigenvalues.sum() if eigenvalues.sum() > 0 else eigenvalues,
            'method': 'RDA'
        }

        return results

    def redundancy_analysis(self, species_data: pd.DataFrame, env_data: pd.DataFrame) -> Dict[str, Any]:
        """
        Redundancy Analysis (RDA) - alias for rda_analysis.

        This method is kept for backward compatibility.
        Use rda_analysis() for new code.
        """
        return self.rda_analysis(species_data, env_data)

    def pcoa_analysis(self, data: pd.DataFrame,
                      distance_metric: str = 'bray_curtis') -> Dict[str, Any]:
        """
        Principal Coordinates Analysis (PCoA).
        
        Parameters:
        -----------
        data : pd.DataFrame
            Species abundance matrix
        distance_metric : str
            Distance metric to use
            
        Returns:
        --------
        dict
            PCoA results
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        if distance_metric in self.distance_metrics:
            distances = self.distance_metrics[distance_metric](data.values)
        else:
            distances = pdist(data.values, metric=distance_metric)
        
        distance_matrix = squareform(distances)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        n = len(distance_matrix)
        H = np.eye(n) - np.ones((n, n)) / n
        B = -0.5 * H @ (distance_matrix**2) @ H
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        eigenvalues, eigenvectors = eig(B)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        idx = np.argsort(eigenvalues)[::-1]
        eigenvalues = eigenvalues[idx].real
        eigenvectors = eigenvectors[:, idx].real
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        positive_eigenvalues = eigenvalues > 1e-8
        eigenvalues = eigenvalues[positive_eigenvalues]
        eigenvectors = eigenvectors[:, positive_eigenvalues]
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        coordinates = eigenvectors @ np.diag(np.sqrt(eigenvalues))
        
        results = {
            'coordinates': pd.DataFrame(
                coordinates,
                index=data.index,
                columns=[f'PCo{i+1}' for i in range(coordinates.shape[1])]
            ),
            'eigenvalues': eigenvalues,
            'explained_variance_ratio': eigenvalues / eigenvalues.sum() if eigenvalues.sum() > 0 else eigenvalues,
            'distance_matrix': distance_matrix,
            'distance_metric': distance_metric,
            'method': 'PCoA'
        }
        
        return results

    def principal_coordinates_analysis(self, data: pd.DataFrame, distance_metric: str = 'bray_curtis') -> Dict[str, Any]:
        """
        Principal Coordinates Analysis (PCoA) - alias for pcoa_analysis.

        This method is kept for backward compatibility.
        Use pcoa_analysis() for new code.
        """
        return self.pcoa_analysis(data, distance_metric)

    # Copyright (c) 2025 Mohamed Z. Hatim
    # Copyright (c) 2025 Mohamed Z. Hatim
    # Copyright (c) 2025 Mohamed Z. Hatim
    
    def _bray_curtis_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Bray-Curtis distance."""
        n_samples = data.shape[0]
        distances = []
        
        for i in range(n_samples):
            for j in range(i + 1, n_samples):
                numerator = np.sum(np.abs(data[i] - data[j]))
                denominator = np.sum(data[i] + data[j])
                
                if denominator == 0:
                    distance = 0
                else:
                    distance = numerator / denominator
                
                distances.append(distance)
        
        return np.array(distances)
    
    def _jaccard_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Jaccard distance."""
    # Copyright (c) 2025 Mohamed Z. Hatim
        binary_data = (data > 0).astype(int)
        
        n_samples = binary_data.shape[0]
        distances = []
        
        for i in range(n_samples):
            for j in range(i + 1, n_samples):
                intersection = np.sum(binary_data[i] * binary_data[j])
                union = np.sum((binary_data[i] + binary_data[j]) > 0)
                
                if union == 0:
                    distance = 0
                else:
                    distance = 1 - (intersection / union)
                
                distances.append(distance)
        
        return np.array(distances)
    
    def _sorensen_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Sorensen distance."""
        binary_data = (data > 0).astype(int)
        
        n_samples = binary_data.shape[0]
        distances = []
        
        for i in range(n_samples):
            for j in range(i + 1, n_samples):
                intersection = np.sum(binary_data[i] * binary_data[j])
                sum_both = np.sum(binary_data[i]) + np.sum(binary_data[j])
                
                if sum_both == 0:
                    distance = 0
                else:
                    distance = 1 - (2 * intersection / sum_both)
                
                distances.append(distance)
        
        return np.array(distances)
    
    def _euclidean_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Euclidean distance."""
        return pdist(data, metric='euclidean')
    
    def _manhattan_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Manhattan distance."""
        return pdist(data, metric='manhattan')
    
    def _canberra_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Canberra distance."""
        return pdist(data, metric='canberra')
    
    def _chord_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate chord distance."""
    # Copyright (c) 2025 Mohamed Z. Hatim
        norms = np.sqrt(np.sum(data**2, axis=1))
        norms[norms == 0] = 1  # Avoid division by zero
        normalized_data = data / norms[:, np.newaxis]
        
        return pdist(normalized_data, metric='euclidean')
    
    def _hellinger_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Hellinger distance."""
    # Copyright (c) 2025 Mohamed Z. Hatim
        row_sums = np.sum(data, axis=1)
        row_sums[row_sums == 0] = 1
        proportions = data / row_sums[:, np.newaxis]
        hellinger_data = np.sqrt(proportions)
        
        return pdist(hellinger_data, metric='euclidean')
    
    # Copyright (c) 2025 Mohamed Z. Hatim
    # Copyright (c) 2025 Mohamed Z. Hatim
    # Copyright (c) 2025 Mohamed Z. Hatim
    
    def _calculate_gradient_lengths(self, site_scores: np.ndarray) -> np.ndarray:
        """Calculate gradient lengths for DCA axes."""
        gradient_lengths = []
        
        for axis in range(site_scores.shape[1]):
            axis_scores = site_scores[:, axis]
            gradient_length = 4 * axis_scores.std()  # Rough approximation
            gradient_lengths.append(gradient_length)
        
        return np.array(gradient_lengths)
    
    def procrustes_analysis(self, data1: np.ndarray, data2: np.ndarray) -> Dict[str, Any]:
        """
        Procrustes analysis for comparing ordinations.
        
        Parameters:
        -----------
        data1 : np.ndarray
            First ordination matrix
        data2 : np.ndarray
            Second ordination matrix
            
        Returns:
        --------
        dict
            Procrustes results
        """
        from scipy.spatial.distance import procrustes
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        mtx1, mtx2, disparity = procrustes(data1, data2)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        correlation = np.corrcoef(mtx1.flatten(), mtx2.flatten())[0, 1]
        
        results = {
            'transformed_data1': mtx1,
            'transformed_data2': mtx2,
            'procrustes_disparity': disparity,
            'correlation': correlation,
            'sum_of_squares': np.sum((mtx1 - mtx2)**2)
        }
        
        return results
    
    def environmental_fitting(self, ordination_scores: pd.DataFrame,
                            env_data: pd.DataFrame,
                            method: str = 'vector') -> Dict[str, Any]:
        """
        Fit environmental vectors or surfaces to ordination.
        
        Parameters:
        -----------
        ordination_scores : pd.DataFrame
            Ordination scores
        env_data : pd.DataFrame
            Environmental variables
        method : str
            Fitting method ('vector' or 'surface')
            
        Returns:
        --------
        dict
            Environmental fitting results
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        common_index = ordination_scores.index.intersection(env_data.index)
        scores_aligned = ordination_scores.loc[common_index]
        env_aligned = env_data.loc[common_index]
        
        results = {
            'method': method,
            'environmental_vectors': {},
            'r_squared': {},
            'p_values': {}
        }
        
        if method == 'vector':
    # Copyright (c) 2025 Mohamed Z. Hatim
            for env_var in env_aligned.columns:
                if env_aligned[env_var].notna().sum() < 3:
                    continue
                
    # Copyright (c) 2025 Mohamed Z. Hatim
                valid_data = env_aligned[env_var].notna()
                
                if valid_data.sum() < 3:
                    continue
                
                X = scores_aligned.loc[valid_data].values
                y = env_aligned.loc[valid_data, env_var].values
                
    # Copyright (c) 2025 Mohamed Z. Hatim
                try:
                    from sklearn.linear_model import LinearRegression
                    
                    reg = LinearRegression()
                    reg.fit(X, y)
                    
    # Copyright (c) 2025 Mohamed Z. Hatim
                    r_squared = reg.score(X, y)
                    
    # Copyright (c) 2025 Mohamed Z. Hatim
                    direction_cosines = reg.coef_
                    
                    results['environmental_vectors'][env_var] = direction_cosines
                    results['r_squared'][env_var] = r_squared
                    
    # Copyright (c) 2025 Mohamed Z. Hatim
                    n_samples = len(y)
                    n_axes = X.shape[1]
                    
                    if n_samples > n_axes + 1:
                        f_stat = (r_squared * (n_samples - n_axes - 1)) / ((1 - r_squared) * n_axes)
                        p_value = 1 - stats.f.cdf(f_stat, n_axes, n_samples - n_axes - 1)
                        results['p_values'][env_var] = p_value
                
                except ImportError:
    # Copyright (c) 2025 Mohamed Z. Hatim
                    correlation = np.corrcoef(X[:, 0], y)[0, 1] if X.shape[1] > 0 else 0
                    results['environmental_vectors'][env_var] = [correlation, 0]
                    results['r_squared'][env_var] = correlation**2
                    results['p_values'][env_var] = 0.05  # Placeholder
        
        return results
    
    def goodness_of_fit_test(self, ordination_results: Dict[str, Any],
                           original_data: pd.DataFrame,
                           distance_metric: str = 'bray_curtis') -> Dict[str, Any]:
        """
        Test goodness of fit for ordination.
        
        Parameters:
        -----------
        ordination_results : dict
            Results from ordination analysis
        original_data : pd.DataFrame
            Original data matrix
        distance_metric : str
            Distance metric for comparison
            
        Returns:
        --------
        dict
            Goodness of fit statistics
        """
    # Copyright (c) 2025 Mohamed Z. Hatim
        if 'site_scores' in ordination_results:
            ord_scores = ordination_results['site_scores']
        elif 'scores' in ordination_results:
            ord_scores = ordination_results['scores']
        elif 'coordinates' in ordination_results:
            ord_scores = ordination_results['coordinates']
        else:
            raise ValueError("No ordination scores found in results")
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        if distance_metric in self.distance_metrics:
            orig_distances = self.distance_metrics[distance_metric](original_data.values)
        else:
            orig_distances = pdist(original_data.values, metric=distance_metric)
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        ord_distances = pdist(ord_scores.values, metric='euclidean')
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        correlation = np.corrcoef(orig_distances, ord_distances)[0, 1]
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        stress = np.sqrt(np.sum((orig_distances - ord_distances)**2) / np.sum(ord_distances**2))
        
    # Copyright (c) 2025 Mohamed Z. Hatim
        shepard_data = pd.DataFrame({
            'original_distance': orig_distances,
            'ordination_distance': ord_distances
        })
        
        results = {
            'correlation': correlation,
            'stress': stress,
            'r_squared': correlation**2,
            'shepard_plot_data': shepard_data,
            'distance_metric': distance_metric
        }
        
        return results


"""
Statistical analysis module for multivariate ecological statistics.

Copyright (c) 2025 Mohamed Z. Hatim
"""

import numpy as np
import pandas as pd
from typing import Union, List, Dict, Tuple, Optional, Any
from scipy import stats
from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csgraph
import warnings
from itertools import combinations


class EcologicalStatistics:
    """Comprehensive statistical analysis for ecological data."""
    
    def __init__(self):
        """Initialize statistical analyzer."""
        self.available_tests = [
            'permanova', 'anosim', 'mrpp', 'mantel',
            'partial_mantel', 'bioenv', 'simper'
        ]
        
        self.distance_metrics = {
            'bray_curtis': self._bray_curtis_distance,
            'jaccard': self._jaccard_distance,
            'euclidean': self._euclidean_distance,
            'manhattan': self._manhattan_distance
        }
    
    def calculate_distance_matrix(self, data: pd.DataFrame, 
                                 metric: str = 'bray_curtis') -> pd.DataFrame:
        """
        Calculate distance matrix using specified metric.
        
        Parameters:
        -----------
        data : pd.DataFrame
            Data matrix (samples x features)
        metric : str
            Distance metric to use
            
        Returns:
        --------
        pd.DataFrame
            Distance matrix
        """
        if metric in self.distance_metrics:
            distances = self.distance_metrics[metric](data.values)
        else:
# Copyright (c) 2025 Mohamed Z. Hatim
            try:
# Copyright (c) 2025 Mohamed Z. Hatim
                scipy_metric = metric.replace('_', '')
                if scipy_metric == 'braycurtis':
                    scipy_metric = 'braycurtis'
                distances = pdist(data.values, metric=scipy_metric)
            except ValueError:
# Copyright (c) 2025 Mohamed Z. Hatim
                distances = self._bray_curtis_distance(data.values)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        distance_matrix = squareform(distances)
        
        return pd.DataFrame(
            distance_matrix,
            index=data.index,
            columns=data.index
        )
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def permanova(self, distance_matrix: Union[pd.DataFrame, np.ndarray],
                  groups: Union[pd.Series, List],
                  permutations: int = 999) -> Dict[str, Any]:
        """
        Permutational Multivariate Analysis of Variance (PERMANOVA).
        
        Parameters:
        -----------
        distance_matrix : pd.DataFrame or np.ndarray
            Distance matrix or data matrix
        groups : pd.Series or list
            Group assignments
        permutations : int
            Number of permutations
            
        Returns:
        --------
        dict
            PERMANOVA results including F-statistic and p-value
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        if isinstance(distance_matrix, pd.DataFrame):
            if distance_matrix.shape[0] == distance_matrix.shape[1]:
# Copyright (c) 2025 Mohamed Z. Hatim
                dist_matrix = distance_matrix.values
            else:
# Copyright (c) 2025 Mohamed Z. Hatim
                dist_matrix = squareform(pdist(distance_matrix.values))
        else:
            if distance_matrix.shape[0] == distance_matrix.shape[1]:
                dist_matrix = distance_matrix
            else:
                dist_matrix = squareform(pdist(distance_matrix))
        
        if isinstance(groups, pd.Series):
            group_labels = groups.values
        else:
            group_labels = np.array(groups)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        observed_f = self._calculate_permanova_f(dist_matrix, group_labels)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_f_stats = []
        n_samples = len(group_labels)
        
        for _ in range(permutations):
# Copyright (c) 2025 Mohamed Z. Hatim
            permuted_groups = np.random.permutation(group_labels)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            f_stat = self._calculate_permanova_f(dist_matrix, permuted_groups)
            permuted_f_stats.append(f_stat)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        p_value = (np.sum(np.array(permuted_f_stats) >= observed_f) + 1) / (permutations + 1)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        unique_groups = np.unique(group_labels)
        df_between = len(unique_groups) - 1
        df_within = n_samples - len(unique_groups)
        df_total = n_samples - 1
        
# Copyright (c) 2025 Mohamed Z. Hatim
        ss_total = self._calculate_total_sum_squares(dist_matrix)
        ss_between = self._calculate_between_sum_squares(dist_matrix, group_labels)
        ss_within = ss_total - ss_between
        
        r_squared = ss_between / ss_total if ss_total > 0 else 0
        
        results = {
            'f_statistic': observed_f,
            'p_value': p_value,
            'r_squared': r_squared,
            'df_between': df_between,
            'df_within': df_within,
            'df_total': df_total,
            'ss_between': ss_between,
            'ss_within': ss_within,
            'ss_total': ss_total,
            'permutations': permutations,
            'method': 'PERMANOVA'
        }
        
        return results
    
    def _calculate_permanova_f(self, dist_matrix: np.ndarray, 
                             group_labels: np.ndarray) -> float:
        """Calculate PERMANOVA F-statistic."""
        n_samples = len(group_labels)
        unique_groups = np.unique(group_labels)
        n_groups = len(unique_groups)
        
        if n_groups < 2:
            return 0.0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        ss_total = self._calculate_total_sum_squares(dist_matrix)
        ss_between = self._calculate_between_sum_squares(dist_matrix, group_labels)
        ss_within = ss_total - ss_between
        
# Copyright (c) 2025 Mohamed Z. Hatim
        df_between = n_groups - 1
        df_within = n_samples - n_groups
        
        if df_within <= 0 or ss_within <= 0:
            return 0.0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        ms_between = ss_between / df_between
        ms_within = ss_within / df_within
        
        f_stat = ms_between / ms_within
        
        return f_stat
    
    def _calculate_total_sum_squares(self, dist_matrix: np.ndarray) -> float:
        """Calculate total sum of squares."""
        n = len(dist_matrix)
        return np.sum(dist_matrix**2) / (2 * n)
    
    def _calculate_between_sum_squares(self, dist_matrix: np.ndarray,
                                     group_labels: np.ndarray) -> float:
        """Calculate between-group sum of squares."""
        unique_groups = np.unique(group_labels)
        n_total = len(group_labels)
        
        ss_between = 0
        
        for group in unique_groups:
            group_mask = group_labels == group
            n_group = np.sum(group_mask)
            
            if n_group <= 1:
                continue
            
# Copyright (c) 2025 Mohamed Z. Hatim
            group_distances = dist_matrix[np.ix_(group_mask, group_mask)]
            ss_group = np.sum(group_distances**2) / (2 * n_group)
            
            ss_between += n_group * ss_group / n_total
        
        ss_total = self._calculate_total_sum_squares(dist_matrix)
        
        return ss_total - ss_between
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def anosim(self, distance_matrix: Union[pd.DataFrame, np.ndarray],
               groups: Union[pd.Series, List],
               permutations: int = 999) -> Dict[str, Any]:
        """
        Analysis of Similarities (ANOSIM).
        
        Parameters:
        -----------
        distance_matrix : pd.DataFrame or np.ndarray
            Distance matrix
        groups : pd.Series or list
            Group assignments
        permutations : int
            Number of permutations
            
        Returns:
        --------
        dict
            ANOSIM results including R-statistic and p-value
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        if isinstance(distance_matrix, pd.DataFrame):
            dist_matrix = distance_matrix.values
        else:
            dist_matrix = distance_matrix
        
        if isinstance(groups, pd.Series):
            group_labels = groups.values
        else:
            group_labels = np.array(groups)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        observed_r = self._calculate_anosim_r(dist_matrix, group_labels)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_r_stats = []
        
        for _ in range(permutations):
            permuted_groups = np.random.permutation(group_labels)
            r_stat = self._calculate_anosim_r(dist_matrix, permuted_groups)
            permuted_r_stats.append(r_stat)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        p_value = (np.sum(np.array(permuted_r_stats) >= observed_r) + 1) / (permutations + 1)
        
        results = {
            'r_statistic': observed_r,
            'p_value': p_value,
            'permutations': permutations,
            'method': 'ANOSIM'
        }
        
        return results
    
    def _calculate_anosim_r(self, dist_matrix: np.ndarray,
                          group_labels: np.ndarray) -> float:
        """Calculate ANOSIM R-statistic."""
        unique_groups = np.unique(group_labels)
        
        if len(unique_groups) < 2:
            return 0.0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        n = len(dist_matrix)
        ranks = np.zeros_like(dist_matrix)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        triu_indices = np.triu_indices(n, k=1)
        distances = dist_matrix[triu_indices]
        distance_ranks = stats.rankdata(distances)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        rank_idx = 0
        for i in range(n):
            for j in range(i + 1, n):
                ranks[i, j] = ranks[j, i] = distance_ranks[rank_idx]
                rank_idx += 1
        
# Copyright (c) 2025 Mohamed Z. Hatim
        within_ranks = []
        between_ranks = []
        
        for i in range(n):
            for j in range(i + 1, n):
                if group_labels[i] == group_labels[j]:
                    within_ranks.append(ranks[i, j])
                else:
                    between_ranks.append(ranks[i, j])
        
        if len(within_ranks) == 0 or len(between_ranks) == 0:
            return 0.0
        
        mean_within = np.mean(within_ranks)
        mean_between = np.mean(between_ranks)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        n_comparisons = len(distances)
        mean_all_ranks = (n_comparisons + 1) / 2
        
        r_stat = (mean_between - mean_within) / (2 * mean_all_ranks - mean_within - mean_between)
        
        return r_stat
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def mrpp(self, distance_matrix: Union[pd.DataFrame, np.ndarray],
             groups: Union[pd.Series, List],
             permutations: int = 999) -> Dict[str, Any]:
        """
        Multi-Response Permutation Procedures (MRPP).
        
        Parameters:
        -----------
        distance_matrix : pd.DataFrame or np.ndarray
            Distance matrix
        groups : pd.Series or list
            Group assignments
        permutations : int
            Number of permutations
            
        Returns:
        --------
        dict
            MRPP results including delta and A statistics
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        if isinstance(distance_matrix, pd.DataFrame):
            dist_matrix = distance_matrix.values
        else:
            dist_matrix = distance_matrix
        
        if isinstance(groups, pd.Series):
            group_labels = groups.values
        else:
            group_labels = np.array(groups)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        observed_delta = self._calculate_mrpp_delta(dist_matrix, group_labels)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        n = len(group_labels)
        all_distances = []
        for i in range(n):
            for j in range(i + 1, n):
                all_distances.append(dist_matrix[i, j])
        
        expected_delta = np.mean(all_distances)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_deltas = []
        
        for _ in range(permutations):
            permuted_groups = np.random.permutation(group_labels)
            delta = self._calculate_mrpp_delta(dist_matrix, permuted_groups)
            permuted_deltas.append(delta)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        p_value = (np.sum(np.array(permuted_deltas) <= observed_delta) + 1) / (permutations + 1)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        a_statistic = (expected_delta - observed_delta) / expected_delta
        
        results = {
            'delta': observed_delta,
            'expected_delta': expected_delta,
            'a_statistic': a_statistic,
            'p_value': p_value,
            'permutations': permutations,
            'method': 'MRPP'
        }
        
        return results
    
    def _calculate_mrpp_delta(self, dist_matrix: np.ndarray,
                            group_labels: np.ndarray) -> float:
        """Calculate MRPP delta statistic."""
        unique_groups, group_counts = np.unique(group_labels, return_counts=True)
        n_total = len(group_labels)
        
        weighted_within_sum = 0
        
        for group, count in zip(unique_groups, group_counts):
            if count <= 1:
                continue
            
            group_mask = group_labels == group
            group_indices = np.where(group_mask)[0]
            
# Copyright (c) 2025 Mohamed Z. Hatim
            within_distances = []
            for i in range(len(group_indices)):
                for j in range(i + 1, len(group_indices)):
                    within_distances.append(dist_matrix[group_indices[i], group_indices[j]])
            
            if within_distances:
                mean_within_distance = np.mean(within_distances)
                weight = count / n_total
                weighted_within_sum += weight * mean_within_distance
        
        return weighted_within_sum
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def mantel_test(self, matrix1: Union[pd.DataFrame, np.ndarray],
                    matrix2: Union[pd.DataFrame, np.ndarray],
                    permutations: int = 999,
                    method: str = 'pearson') -> Dict[str, Any]:
        """
        Mantel test for matrix correlation.
        
        Parameters:
        -----------
        matrix1 : pd.DataFrame or np.ndarray
            First matrix
        matrix2 : pd.DataFrame or np.ndarray
            Second matrix
        permutations : int
            Number of permutations
        method : str
            Correlation method ('pearson', 'spearman', 'kendall')
            
        Returns:
        --------
        dict
            Mantel test results
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        if isinstance(matrix1, pd.DataFrame):
            mat1 = matrix1.values
        else:
            mat1 = matrix1
        
        if isinstance(matrix2, pd.DataFrame):
            mat2 = matrix2.values
        else:
            mat2 = matrix2
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if mat1.shape != mat2.shape:
            raise ValueError("Matrices must have the same dimensions")
        
# Copyright (c) 2025 Mohamed Z. Hatim
        n = mat1.shape[0]
        triu_indices = np.triu_indices(n, k=1)
        
        vec1 = mat1[triu_indices]
        vec2 = mat2[triu_indices]
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if method == 'pearson':
            observed_r = np.corrcoef(vec1, vec2)[0, 1]
        elif method == 'spearman':
            observed_r = stats.spearmanr(vec1, vec2)[0]
        elif method == 'kendall':
            observed_r = stats.kendalltau(vec1, vec2)[0]
        else:
            raise ValueError(f"Unknown correlation method: {method}")
        
# Copyright (c) 2025 Mohamed Z. Hatim
        if np.isnan(observed_r):
            observed_r = 0.0
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_correlations = []
        
        for _ in range(permutations):
# Copyright (c) 2025 Mohamed Z. Hatim
            perm_indices = np.random.permutation(n)
            mat2_permuted = mat2[np.ix_(perm_indices, perm_indices)]
            
# Copyright (c) 2025 Mohamed Z. Hatim
            vec2_permuted = mat2_permuted[triu_indices]
            
# Copyright (c) 2025 Mohamed Z. Hatim
            if method == 'pearson':
                r = np.corrcoef(vec1, vec2_permuted)[0, 1]
            elif method == 'spearman':
                r = stats.spearmanr(vec1, vec2_permuted)[0]
            elif method == 'kendall':
                r = stats.kendalltau(vec1, vec2_permuted)[0]
            
            if np.isnan(r):
                r = 0.0
            
            permuted_correlations.append(r)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_correlations = np.array(permuted_correlations)
        p_value = (np.sum(np.abs(permuted_correlations) >= np.abs(observed_r)) + 1) / (permutations + 1)
        
        results = {
            'correlation': observed_r,
            'p_value': p_value,
            'permutations': permutations,
            'method': f'Mantel_{method}',
            'permuted_correlations': permuted_correlations
        }
        
        return results
    
    def partial_mantel_test(self, matrix1: Union[pd.DataFrame, np.ndarray],
                           matrix2: Union[pd.DataFrame, np.ndarray],
                           matrix3: Union[pd.DataFrame, np.ndarray],
                           permutations: int = 999,
                           method: str = 'pearson') -> Dict[str, Any]:
        """
        Partial Mantel test controlling for a third matrix.
        
        Parameters:
        -----------
        matrix1 : pd.DataFrame or np.ndarray
            First matrix
        matrix2 : pd.DataFrame or np.ndarray
            Second matrix
        matrix3 : pd.DataFrame or np.ndarray
            Control matrix
        permutations : int
            Number of permutations
        method : str
            Correlation method
            
        Returns:
        --------
        dict
            Partial Mantel test results
        """
# Copyright (c) 2025 Mohamed Z. Hatim
        def extract_upper_tri(matrix):
            if isinstance(matrix, pd.DataFrame):
                mat = matrix.values
            else:
                mat = matrix
            
            n = mat.shape[0]
            triu_indices = np.triu_indices(n, k=1)
            return mat[triu_indices]
        
        vec1 = extract_upper_tri(matrix1)
        vec2 = extract_upper_tri(matrix2)
        vec3 = extract_upper_tri(matrix3)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        def partial_correlation(x, y, z):
            """Calculate partial correlation between x and y controlling for z."""
# Copyright (c) 2025 Mohamed Z. Hatim
            if method.lower() == 'spearman':
                rxy = stats.spearmanr(x, y)[0]
                rxz = stats.spearmanr(x, z)[0]
                ryz = stats.spearmanr(y, z)[0]
            else:
                rxy = np.corrcoef(x, y)[0, 1]
                rxz = np.corrcoef(x, z)[0, 1]
                ryz = np.corrcoef(y, z)[0, 1]

# Copyright (c) 2025 Mohamed Z. Hatim
            rxy = 0.0 if np.isnan(rxy) else rxy
            rxz = 0.0 if np.isnan(rxz) else rxz
            ryz = 0.0 if np.isnan(ryz) else ryz

# Copyright (c) 2025 Mohamed Z. Hatim
            denominator = np.sqrt((1 - rxz**2) * (1 - ryz**2))

            if denominator == 0:
                return 0.0

            partial_r = (rxy - rxz * ryz) / denominator

            return partial_r
        
# Copyright (c) 2025 Mohamed Z. Hatim
        observed_partial_r = partial_correlation(vec1, vec2, vec3)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_partial_correlations = []
        n = matrix1.shape[0] if isinstance(matrix1, np.ndarray) else matrix1.shape[0]
        
        for _ in range(permutations):
# Copyright (c) 2025 Mohamed Z. Hatim
            perm_indices = np.random.permutation(n)
            
            if isinstance(matrix2, pd.DataFrame):
                mat2_permuted = matrix2.iloc[perm_indices, perm_indices].values
            else:
                mat2_permuted = matrix2[np.ix_(perm_indices, perm_indices)]
            
            vec2_permuted = extract_upper_tri(mat2_permuted)
            
            partial_r = partial_correlation(vec1, vec2_permuted, vec3)
            permuted_partial_correlations.append(partial_r)
        
# Copyright (c) 2025 Mohamed Z. Hatim
        permuted_partial_correlations = np.array(permuted_partial_correlations)
        p_value = (np.sum(np.abs(permuted_partial_correlations) >= np.abs(observed_partial_r)) + 1) / (permutations + 1)
        
        results = {
            'partial_correlation': observed_partial_r,
            'p_value': p_value,
            'permutations': permutations,
            'method': f'Partial_Mantel_{method}',
            'permuted_correlations': permuted_partial_correlations
        }
        
        return results
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def indicator_species_analysis(self, species_data: pd.DataFrame,
                                 groups: Union[pd.Series, List],
                                 permutations: int = 999) -> Dict[str, Any]:
        """
        Indicator Species Analysis (IndVal).
        
        Parameters:
        -----------
        species_data : pd.DataFrame
            Species abundance matrix
        groups : pd.Series or list
            Group assignments
        permutations : int
            Number of permutations
            
        Returns:
        --------
        dict
            IndVal results for each species
        """
        if isinstance(groups, pd.Series):
            group_labels = groups.values
        else:
            group_labels = np.array(groups)
        
        unique_groups = np.unique(group_labels)
        results = {}
        
        for species in species_data.columns:
            species_abundances = species_data[species].values
            
# Copyright (c) 2025 Mohamed Z. Hatim
            group_indvals = {}
            
            for group in unique_groups:
                group_mask = group_labels == group
                
# Copyright (c) 2025 Mohamed Z. Hatim
                group_abundance = species_abundances[group_mask]
                total_abundance = species_abundances.sum()
                
                if total_abundance == 0:
                    relative_abundance = 0
                else:
                    relative_abundance = group_abundance.sum() / total_abundance
                
# Copyright (c) 2025 Mohamed Z. Hatim
                group_presence = (group_abundance > 0).sum()
                group_size = group_mask.sum()
                
                if group_size == 0:
                    relative_frequency = 0
                else:
                    relative_frequency = group_presence / group_size
                
# Copyright (c) 2025 Mohamed Z. Hatim
                indval = relative_abundance * relative_frequency * 100
                
                group_indvals[group] = {
                    'indval': indval,
                    'relative_abundance': relative_abundance,
                    'relative_frequency': relative_frequency
                }
            
# Copyright (c) 2025 Mohamed Z. Hatim
            max_group = max(group_indvals.keys(), key=lambda g: group_indvals[g]['indval'])
            max_indval = group_indvals[max_group]['indval']
            
# Copyright (c) 2025 Mohamed Z. Hatim
            permuted_indvals = []

            for _ in range(permutations):
                permuted_groups = np.random.permutation(group_labels)

# Copyright (c) 2025 Mohamed Z. Hatim
                perm_max_indval = 0
                for perm_group in unique_groups:
                    perm_group_mask = permuted_groups == perm_group
                    perm_group_abundance = species_abundances[perm_group_mask]

                    if total_abundance == 0:
                        perm_rel_abundance = 0
                    else:
                        perm_rel_abundance = perm_group_abundance.sum() / total_abundance

                    perm_group_presence = (perm_group_abundance > 0).sum()
                    perm_group_size = perm_group_mask.sum()

                    if perm_group_size == 0:
                        perm_rel_frequency = 0
                    else:
                        perm_rel_frequency = perm_group_presence / perm_group_size

                    perm_indval = perm_rel_abundance * perm_rel_frequency * 100
                    if perm_indval > perm_max_indval:
                        perm_max_indval = perm_indval

                permuted_indvals.append(perm_max_indval)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            p_value = (np.sum(np.array(permuted_indvals) >= max_indval) + 1) / (permutations + 1)
            
            results[species] = {
                'max_group': max_group,
                'indval': max_indval,
                'p_value': p_value,
                'group_details': group_indvals
            }
        
        return results
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def simper_analysis(self, species_data: pd.DataFrame,
                       groups: Union[pd.Series, List],
                       distance_metric: str = 'bray_curtis') -> Dict[str, Any]:
        """
        Similarity Percentages (SIMPER) analysis.
        
        Parameters:
        -----------
        species_data : pd.DataFrame
            Species abundance matrix
        groups : pd.Series or list
            Group assignments
        distance_metric : str
            Distance metric to use
            
        Returns:
        --------
        dict
            SIMPER results
        """
        if isinstance(groups, pd.Series):
            group_labels = groups.values
        else:
            group_labels = np.array(groups)
        
        unique_groups = np.unique(group_labels)
        results = {}
        
# Copyright (c) 2025 Mohamed Z. Hatim
        for group in unique_groups:
            group_mask = group_labels == group
            group_data = species_data[group_mask]
            
            if len(group_data) < 2:
                continue
            
# Copyright (c) 2025 Mohamed Z. Hatim
            similarities = []
            species_contributions = {species: [] for species in species_data.columns}
            
            for i in range(len(group_data)):
                for j in range(i + 1, len(group_data)):
# Copyright (c) 2025 Mohamed Z. Hatim
                    sample1 = group_data.iloc[i].values
                    sample2 = group_data.iloc[j].values
                    
                    if distance_metric == 'bray_curtis':
                        distance = self._bray_curtis_single(sample1, sample2)
                    else:
                        distance = np.linalg.norm(sample1 - sample2)
                    
                    similarity = 1 - distance
                    similarities.append(similarity)
                    
# Copyright (c) 2025 Mohamed Z. Hatim
                    for k, species in enumerate(species_data.columns):
# Copyright (c) 2025 Mohamed Z. Hatim
                        total_sum = sample1.sum() + sample2.sum()
                        if total_sum > 0:
                            species_contribution = abs(sample1[k] - sample2[k]) / total_sum
                        else:
                            species_contribution = 0
                        species_contributions[species].append(species_contribution)
            
# Copyright (c) 2025 Mohamed Z. Hatim
            avg_similarity = np.mean(similarities) if similarities else 0
            
            species_avg_contrib = {}
            for species in species_data.columns:
                if species_contributions[species]:
                    species_avg_contrib[species] = np.mean(species_contributions[species])
                else:
                    species_avg_contrib[species] = 0
            
            results[f'within_{group}'] = {
                'average_similarity': avg_similarity,
                'species_contributions': species_avg_contrib
            }
        
# Copyright (c) 2025 Mohamed Z. Hatim
        for i, group1 in enumerate(unique_groups):
            for group2 in unique_groups[i + 1:]:
                group1_mask = group_labels == group1
                group2_mask = group_labels == group2
                
                group1_data = species_data[group1_mask]
                group2_data = species_data[group2_mask]
                
# Copyright (c) 2025 Mohamed Z. Hatim
                dissimilarities = []
                species_contributions = {species: [] for species in species_data.columns}
                
                for idx1, (_, sample1) in enumerate(group1_data.iterrows()):
                    for idx2, (_, sample2) in enumerate(group2_data.iterrows()):
                        s1_values = sample1.values
                        s2_values = sample2.values
                        
                        if distance_metric == 'bray_curtis':
                            dissimilarity = self._bray_curtis_single(s1_values, s2_values)
                        else:
                            dissimilarity = np.linalg.norm(s1_values - s2_values)
                        
                        dissimilarities.append(dissimilarity)
                        
# Copyright (c) 2025 Mohamed Z. Hatim
                        for k, species in enumerate(species_data.columns):
                            total_sum = s1_values.sum() + s2_values.sum()
                            if total_sum > 0:
                                contrib = abs(s1_values[k] - s2_values[k]) / total_sum
                            else:
                                contrib = 0
                            species_contributions[species].append(contrib)
                
# Copyright (c) 2025 Mohamed Z. Hatim
                avg_dissimilarity = np.mean(dissimilarities) if dissimilarities else 0
                
                species_avg_contrib = {}
                for species in species_data.columns:
                    if species_contributions[species]:
                        species_avg_contrib[species] = np.mean(species_contributions[species])
                    else:
                        species_avg_contrib[species] = 0
                
                results[f'between_{group1}_{group2}'] = {
                    'average_dissimilarity': avg_dissimilarity,
                    'species_contributions': species_avg_contrib
                }
        
        return results
    
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
# Copyright (c) 2025 Mohamed Z. Hatim
    
    def _bray_curtis_single(self, x: np.ndarray, y: np.ndarray) -> float:
        """Calculate Bray-Curtis distance between two samples."""
        numerator = np.sum(np.abs(x - y))
        denominator = np.sum(x + y)
        
        if denominator == 0:
            return 0.0
        
        return numerator / denominator
    
    def _bray_curtis_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Bray-Curtis distance matrix."""
        n_samples = data.shape[0]
        distances = []
        
        for i in range(n_samples):
            for j in range(i + 1, n_samples):
                distance = self._bray_curtis_single(data[i], data[j])
                distances.append(distance)
        
        return np.array(distances)
    
    def _jaccard_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Jaccard distance matrix."""
        binary_data = (data > 0).astype(int)
        return pdist(binary_data, metric='jaccard')
    
    def _euclidean_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Euclidean distance matrix."""
        return pdist(data, metric='euclidean')
    
    def _manhattan_distance(self, data: np.ndarray) -> np.ndarray:
        """Calculate Manhattan distance matrix."""
        return pdist(data, metric='manhattan')